"""
"whatrecord parse" is used to parse and interpret any of whatrecord's supported
file formats, dumping the results to the console (standard output) in JSON
format, by default.

Unless ``--disable-macros`` is specified, all text will go through the macro
context as if the files were being loaded in an IOC shell.
"""

import argparse
import json
import logging
from typing import List, Optional

from ..common import AnyPath, FileFormat, IocMetadata
from ..parse import ParseResult, parse
from ..shell import LoadedIoc
from ..util import write_to_file

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        'filename',
        type=str,
        help="Startup script filename"
    )

    parser.add_argument(
        "-if",
        "--input-format",
        type=str,
        required=False,
        help=(
            "The file format.  For files that lack a recognized extension or "
            "are otherwise misidentified by whatrecord."
        ),
    )

    parser.add_argument(
        "--v3",
        action="store_true",
        help="Use V3 database grammar instead of V4 where possible"
    )

    parser.add_argument(
        "--dbd",
        type=str,
        help="The dbd file, if parsing a database",
    )

    parser.add_argument(
        "--standin-directory",
        type=str,
        nargs="*",
        help='Map a "stand-in" directory to another on disk',
    )

    parser.add_argument(
        "--macros",
        type=str,
        help="Macro to add, in the usual form ``macro=value,...``",
    )

    parser.add_argument(
        "--disable-macros",
        action="store_true",
        help="Disable macro handling, leaving unexpanded macros in the output."
    )

    parser.add_argument(
        "-o", "--output",
        type=str,
        required=False,
        help="Output file to write to.  Defaults to standard output.",
    )

    parser.add_argument(
        "-of", "--output-format",
        type=str,
        default="json",
        help=(
            "Defaults to 'json', some input formats may support "
            "'console', which aims to be convenient for console output, "
            "or 'file', which should closely resemble the input file"
        ),
    )

    parser.add_argument(
        '--use-gdb',
        action="store_true",
        help="Use metadata derived from the script binary",
    )

    parser.add_argument(
        "--expand",
        action="store_true",
        help="Expand a substitutions file, as in the msi tool",
    )

    return parser


def parse_from_cli_args(
    filename: AnyPath,
    dbd: Optional[str] = None,
    standin_directory: Optional[List[str]] = None,
    macros: Optional[str] = None,
    disable_macros: bool = False,
    use_gdb: bool = False,
    format: Optional[str] = None,
    expand: bool = False,
    v3: bool = False,
) -> ParseResult:
    """
    Generically parse either a startup script or a database file.

    This variant uses the raw CLI arguments, mapping them on to those that
    `parse` expects.  For programmatic usage, please use ``parse()`` directly.

    Unless ``disable_macros`` is set, all text will go through the macro
    context as if the files were being loaded in an IOC shell.

    Parameters
    ----------
    filename : str or pathlib.Path
        The filename to parse.

    dbd : str or pathlib.Path, optional
        The associated database definition file, if parsing a database file.

    standin_directories : list, optional
        A list of substitute directories of the form ``DirectoryA=DirectoryB``.

    macros : str, optional
        Macro string to use when parsing the file.  Ignored if
        ``disable_macros`` is set.

    disable_macros : bool, optional
        Disable macro handling, leaving unexpanded macros in the output.

    expand : bool, optional
        Expand a substitutions file.

    v3 : bool, optional
        Use V3 database grammar where applicable.
    """
    standin_directories = dict(
        path.split("=", 1) for path in standin_directory or ""
    )

    if isinstance(filename, str) and filename.startswith("{"):   # }
        # TODO - argparse fixup?
        ioc_metadata = IocMetadata.from_dict(json.loads(filename))
        # TODO macros, use_gdb, ...
        # ioc_metadata.macros = macros
        return LoadedIoc.from_metadata(
            ioc_metadata
        )

    try:
        format = FileFormat(format) if format is not None else None
    except ValueError:
        options = [fmt.name for fmt in list(FileFormat)]
        raise ValueError(
            f"{format!r} is not a valid FileFormat. Options include: "
            f"{options}"
        )

    return parse(
        filename,
        dbd=dbd,
        standin_directories=standin_directories,
        macros=macros,
        disable_macros=disable_macros,
        use_gdb=use_gdb,
        format=format,
        expand=expand,
        v3=v3,
    )


def main(
    filename: AnyPath,
    dbd: Optional[str] = None,
    standin_directory: Optional[List[str]] = None,
    input_format: Optional[str] = None,
    output_format: str = "json",
    output: Optional[str] = None,
    macros: Optional[str] = None,
    disable_macros: bool = False,
    use_gdb: bool = False,
    expand: bool = False,
    v3: bool = False,
):
    result = parse_from_cli_args(
        filename=filename,
        dbd=dbd,
        standin_directory=standin_directory,
        macros=macros,
        disable_macros=disable_macros,
        use_gdb=use_gdb,
        format=input_format,
        expand=expand,
        v3=v3,
    )

    write_to_file(result, filename=output, format=output_format)



import pathlib
import pprint

import apischema
import pytest

from ..streamdevice import StreamProtocol
from . import conftest

PROTOCOL_FILES = list((conftest.MODULE_PATH / "iocs").glob("**/*.proto"))

additional_files = conftest.MODULE_PATH / "streamdevice_filenames.txt"
if additional_files.exists():
    for additional in open(additional_files, "rt").read().splitlines():
        PROTOCOL_FILES.append(pathlib.Path(additional))


protocol_files = pytest.mark.parametrize(
    "protocol_file",
    [
        pytest.param(
            protocol_file,
            id="/".join(protocol_file.parts[-2:])
        )
        for protocol_file in PROTOCOL_FILES
    ]
)


@protocol_files
def test_parse(protocol_file):
    proto = StreamProtocol.from_file(protocol_file)

    serialized = apischema.serialize(proto)
    pprint.pprint(serialized)
    apischema.deserialize(StreamProtocol, serialized)



import pathlib
# import pprint
import textwrap

import apischema
import pytest

from ..snl import SequencerProgram
from . import conftest

SNL_FILES = list((conftest.MODULE_PATH / "iocs").glob("**/*.st"))

additional_files = conftest.MODULE_PATH / "sequencer_filenames.txt"
if additional_files.exists():
    for additional in open(additional_files, "rt").read().splitlines():
        SNL_FILES.append(pathlib.Path(additional))


SNL_IDS = [
    "/".join(program_filename.parts[-2:])
    for program_filename in SNL_FILES
]


@pytest.fixture(
    params=SNL_FILES,
    ids=SNL_IDS,
)
def snl_filename(request) -> pathlib.Path:
    return request.param


@pytest.fixture
def snl_program(snl_filename) -> SequencerProgram:
    with open(snl_filename, "rt") as fp:
        source = SequencerProgram.preprocess(
            fp.read(), search_path=snl_filename.parent
        )
    has_program = any(
        line.startswith("program")
        for line in source.splitlines()
    )
    if not has_program:
        raise pytest.skip(
            "No program found in source "
            "(perhaps it's meant to be included from another?)"
        )

    return SequencerProgram.from_file(snl_filename)


def test_parse(snl_program: SequencerProgram):
    # from ..format import FormatContext
    # ctx = FormatContext()
    # print(ctx.render_object(program, "console"))
    print(snl_program)

    serialized = apischema.serialize(snl_program)
    # pprint.pprint(serialized)
    apischema.deserialize(SequencerProgram, serialized)
    # deserialized = apischema.deserialize(SequencerProgram, serialized)
    # assert deserialized == snl_program

    round_trip = str(snl_program)
    print("Round-tripped to")
    print(round_trip)
    assert "(context=(" not in round_trip
    assert "(Tree=(" not in round_trip
    assert "(Token=(" not in round_trip


def test_graph(snl_filename: pathlib.Path, snl_program: SequencerProgram):
    graph = snl_program.as_graph()
    digraph = graph.to_digraph()

    from ..bin.graph import render_graph_to_file
    render_graph_to_file(digraph, filename=f"{snl_filename}.pdf")


@pytest.mark.parametrize(
    "source, partial_error",
    [
        pytest.param(
            """\
            program name
            #define VALUE 5

            double a[VALUE];

            ss simple {
                state simple {
                    when () {} exit
                }
            }
            """,
            "No terminal matches 'V' in the current parser context",
            id="preprocessor_define"
        ),
        pytest.param(
            """\
            program name
            #define DEFINED 1

            #if DEFINED
            #else
            !!! invalid syntax
            #endif
            double a[VALUE];

            ss simple {
                state simple {
                    when () {} exit
                }
            }
            """,
            "No terminal matches '!' in the current parser context",
            id="preprocessor_define"
        ),
    ]
)
@pytest.mark.xfail(strict=True, reason="preprocessor incomplete")
def test_tofix(source, partial_error):
    source = textwrap.dedent(source)
    try:
        SequencerProgram.from_string(source)
    except Exception as ex:
        if partial_error in str(ex):
            raise
        print(ex)



import pprint
import textwrap

import apischema
import pytest

from ..plugins import epicsarch
from . import conftest

ARCH_FILES = list((conftest.MODULE_PATH / "epicsarch").glob("*.txt"))

arch_filenames = pytest.mark.parametrize(
    "filename",
    [
        pytest.param(
            filename,
            id="/".join(filename.parts[-2:])
        )
        for filename in ARCH_FILES
    ]
)


@arch_filenames
def test_parse(filename):
    parsed = epicsarch.LclsEpicsArchFile.from_file(filename)

    serialized = apischema.serialize(parsed)
    pprint.pprint(serialized)
    apischema.deserialize(epicsarch.LclsEpicsArchFile, serialized)


def test_warnings():
    fn = conftest.MODULE_PATH / "epicsarch" / "test.txt"
    parsed = epicsarch.LclsEpicsArchFile.from_file(fn)

    from ..format import FormatContext

    ctx = FormatContext()
    print(ctx.render_object(parsed, "console"))

    types = [warning.type_ for warning in parsed.warnings]

    assert set(types) == {
        "duplicate_alias",
        "duplicate_pv",
        "alias_is_pv",
        "pv_is_alias",
        "missing_file",
        "recursive_include",
    }

    warning_texts = [warning.text for warning in parsed.warnings]
    assert any("Duplicate pvname: pvname5" in text for text in warning_texts)
    assert any("Alias name is a PV: pvname5" in text for text in warning_texts)
    assert any("PV name matches alias: descB" in text for text in warning_texts)
    assert any("missing" in text for text in warning_texts)
    assert any("Recursively included" in text for text in warning_texts)


def test_empty_section():
    epicsarch.LclsEpicsArchFile.from_string(
        textwrap.dedent(
            """\
            #*desc
            #*desc2
            pv
            """
        )
    )



from __future__ import annotations

import logging
import pathlib
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional

import lark
from epicsmacrolib import split_iocsh_line

from . import transformer
from .common import (AnyPath, FullLoadContext, LoadContext, ShellStateHandler,
                     StringWithContext)
from .db import PVAFieldReference, RecordInstance
from .transformer import context_from_token

logger = logging.getLogger(__name__)


def _separate_by_class(items, mapping):
    """Separate ``items`` by type into ``mapping`` of collections."""
    # TODO: should remove this entirely and rewrite the transformer
    for item in items:
        if item == ";":
            continue

        container = mapping[type(item)]
        if isinstance(container, list):
            container.append(item)
        else:
            container[item.name] = item


@dataclass
class ConfigurationSetting:
    name: str
    value: str


@dataclass
class VariableAssignment:
    name: str
    value: str


@dataclass
class Command:
    name: str
    arguments: List[str]


@dataclass
class ProtocolDefinition:
    context: FullLoadContext
    name: str
    handlers: Dict[str, HandlerDefinition] = field(default_factory=dict)
    variables: Dict[str, str] = field(default_factory=dict)
    commands: List[Command] = field(default_factory=list)
    config: Dict[str, str] = field(default_factory=dict)


@dataclass
class HandlerDefinition:
    name: str
    commands: List[Command] = field(default_factory=list)


@dataclass
class StreamProtocol:
    """Representation of a StreamDevice protocol."""
    variables: Dict[str, str] = field(default_factory=dict)
    protocols: Dict[str, ProtocolDefinition] = field(default_factory=dict)
    comments: List[str] = field(default_factory=list)
    config: Dict[str, str] = field(default_factory=dict)
    handlers: Dict[str, HandlerDefinition] = field(default_factory=dict)

    @classmethod
    def from_string(
        cls, contents, filename=None,
    ) -> StreamProtocol:
        """Load a protocol file given its string contents."""
        comments = []
        grammar = lark.Lark.open_from_package(
            "whatrecord",
            "streamdevice.lark",
            search_paths=("grammar", ),
            parser="earley",
            maybe_placeholders=False,
            # lexer_callbacks={"COMMENT": comments.append},
        )

        proto = _ProtocolTransformer(cls, filename).transform(
            grammar.parse(contents)
        )
        proto.comments = comments
        return proto

    @classmethod
    def from_file_obj(cls, fp, filename=None) -> StreamProtocol:
        """Load a protocol file given a file object."""
        return cls.from_string(
            fp.read(),
            filename=getattr(fp, "name", filename),
        )

    @classmethod
    def from_file(cls, fn) -> StreamProtocol:
        """
        Load a StreamDevice protocol file.

        Parameters
        ----------
        filename : pathlib.Path or str
            The filename.

        Returns
        -------
        protocol : StreamProtocol
            The StreamDevice protocol.
        """
        with open(fn, "rt") as fp:
            return cls.from_string(fp.read(), filename=fn)


@lark.visitors.v_args(inline=True)
class _ProtocolTransformer(lark.visitors.Transformer):
    def __init__(self, cls, fn, visit_tokens=False):
        super().__init__(visit_tokens=visit_tokens)
        self.fn = str(fn)
        self.cls = cls

    @lark.visitors.v_args(tree=True)
    def protocol(self, body):
        proto = self.cls()
        variables = []
        config_settings = []
        _separate_by_class(
            body.children,
            {
                VariableAssignment: variables,
                ProtocolDefinition: proto.protocols,
                ConfigurationSetting: config_settings,
                HandlerDefinition: proto.handlers,
            }
        )
        proto.config = {
            config.name: config.value
            for config in config_settings
        }
        proto.variables = {
            var.name: var.value
            for var in variables
        }
        return proto

    def assignment(self, variable, value):
        return VariableAssignment(str(variable), str(value))

    def protocol_body(self, *items):
        return sum((getattr(item, "children", [item]) for item in items), [])

    def standalone_variable_ref(self, ref, *_):
        return Command(
            name=str(ref),
            arguments=[],
        )

    def config_set(self, command: lark.Tree):
        return ConfigurationSetting(
            name=str(command.data),
            value=" ".join(str(child) for child in command.children),
        )

    value_part = transformer.pass_through

    def value(self, *items):
        return [
            str(item) for item in items
            if item is not None
        ]

    def _generic_command(command_name: str):
        def _wrapped(self, *values):
            args = [
                str(val)
                for value in values
                for val in value
            ]
            return Command(
                name=command_name,
                arguments=args,
            )
        return _wrapped

    disconnect = _generic_command("disconnect")
    connect = _generic_command("connect")
    out = _generic_command("out")
    locals()["in"] = _generic_command("in")  # "in" is reserved, oops
    wait = _generic_command("wait")
    event = _generic_command("event")
    exec = _generic_command("exec")

    def integer(self, value):
        return "".join(v for v in value)

    def hex_byte(self, *digits):
        return "0x" + "".join(digits)

    def octal_byte(self, *digits):
        return "".join(digits)

    def number(self, *digits):
        return "".join(digits)

    def user_defined_command(self, name, args=None):
        return Command(
            name=name,
            arguments=list(args) if args else [],
        )

    user_defined_command_args = transformer.pass_through
    command = transformer.pass_through

    def protocol_name(self, name):
        return StringWithContext(
            name,
            context=context_from_token(self.fn, name),
        )

    def protocol_def(self, name, children):
        defn = ProtocolDefinition(
            name=name,
            context=name.context,
        )
        variables = []
        config_settings = []
        _separate_by_class(
            children,
            {
                VariableAssignment: variables,
                HandlerDefinition: defn.handlers,
                Command: defn.commands,
                ConfigurationSetting: config_settings,
            }
        )
        defn.config = {
            config.name: config.value
            for config in config_settings
        }
        defn.variables = {
            var.name: var.value
            for var in variables
        }
        return defn

    def handler_body(self, *items):
        commands = []
        for item in items:
            if isinstance(item, tuple):
                commands.append(item[0])  # (Command, EOL)
            else:
                commands.append(item)
        return commands

    def handler_def(self, name, children, *_):
        defn = HandlerDefinition(
            name=name,
        )
        _separate_by_class(
            children,
            {
                Command: defn.commands,
            }
        )
        return defn


_handler = ShellStateHandler.generic_handler_decorator


@dataclass
class StreamDeviceState(ShellStateHandler):
    """
    StreamDevice IOC shell state handler / container.

    Contains hooks for StreamDevice-related commands and state information.

    Attributes
    ----------
    protocols : Dict[str, StreamProtocol]
        Loaded StreamDevice protocols by name.
    """

    metadata_key: ClassVar[str] = "streamdevice"
    protocols: Dict[str, StreamProtocol] = field(
        default_factory=dict
    )

    def find_streamdevice_protocol(self, filename: AnyPath) -> pathlib.Path:
        shell_state = self.primary_handler
        assert shell_state is not None
        return shell_state._fix_path_with_search_list(
            filename,
            shell_state.paths_from_env_var("STREAM_PROTOCOL_PATH", default=".")
        )

    def load_streamdevice_protocol(
        self,
        filename: AnyPath,
    ) -> StreamProtocol:
        """Load a StreamDevice protocol file."""
        filename = self.find_streamdevice_protocol(filename)
        key = str(filename)
        if key not in self.protocols:
            shell_state = self.primary_handler
            fn, contents = shell_state.load_file(filename)
            self.protocols[key] = StreamProtocol.from_string(
                contents,
                filename=fn
            )
        return self.protocols[key]

    def annotate_record(self, record: RecordInstance) -> Optional[Dict[str, Any]]:
        """Hook to annotate a record after being loaded."""
        dtype = record.fields.get("DTYP", None)
        if not dtype or getattr(dtype, "value", None) != "stream":
            return

        info_field = record.fields.get("INP", record.fields.get("OUT", None))
        if not info_field or isinstance(info_field, PVAFieldReference):
            return {
                "error": "INP/OUT not defined correctly"
            }
        if not isinstance(info_field.value, str):
            return {
                "error": "INP/OUT not defined correctly (JSON)"
            }

        info_field = info_field.value.strip()
        annotation: Dict[str, Any] = {
            "context": [],
        }
        try:
            proto_file, proto_name, *proto_args = split_iocsh_line(info_field).argv
            proto_file = proto_file.lstrip("@ ")
        except Exception:
            annotation["error"] = (
                f"Invalid StreamDevice input/output field: {info_field!r}"
            )
            return annotation

        try:
            protocol = self.load_streamdevice_protocol(proto_file)
        except Exception as ex:
            annotation["error"] = f"{ex.__class__.__name__}: {ex}"
            protocol = None
            return annotation

        if proto_name in protocol.protocols:
            specific_protocol = protocol.protocols[proto_name]
            annotation["context"] = specific_protocol.context
            annotation["protocol"] = specific_protocol
        else:
            annotation["context"] = [LoadContext(proto_file, 0)]
            annotation["error"] = (
                f"Unknown protocol {proto_name!r} in {proto_file}; "
                f"options are: {list(protocol.protocols)}"
            )
        annotation["protocol_name"] = proto_name
        annotation["protocol_args"] = proto_args
        return annotation



from __future__ import annotations

import logging
import sys
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, Generator, List, Optional, Tuple, Union

import apischema

from .. import util
from ..common import (IocMetadata, IocshScript, PVRelations, RecordInstance,
                      RecordInstanceSummary, ScriptPVRelations, WhatRecord)

logger = logging.getLogger(__name__)


class TooManyRecordsError(Exception):
    ...


@dataclass
class PluginResults:
    files_to_monitor: Dict[str, str] = field(default_factory=dict)
    record_to_metadata_keys: Dict[str, List[str]] = field(default_factory=dict)
    metadata_by_key: Dict[str, Any] = field(default_factory=dict)
    metadata: Any = None
    execution_info: Dict[str, Any] = field(default_factory=dict)
    nested: Optional[Dict[str, PluginResults]] = None

    def is_loaded_file(self, fn: str) -> bool:
        """Is the given file one that was loaded in the plugin?"""
        if fn in self.files_to_monitor:
            return True
        return any(
            nested.is_loaded_file(fn)
            for nested in (self.nested or {}).values()
        )

    def find_record_metadata(self, record: str) -> Generator[Any, None, None]:
        """Find all metadata for the given record name."""
        for key in self.record_to_metadata_keys.get(record) or []:
            try:
                yield self.metadata_by_key[key]
            except KeyError:
                logger.debug(
                    "Consistency error in plugin: missing metadata key(s) %r "
                    "for record %s",
                    key, record
                )

        for nested in (self.nested or {}).values():
            yield from nested.find_record_metadata(record)

    def find_by_key(self, key: str) -> Generator[Tuple[str, Any], None, None]:
        """Find all metadata for the given record name."""
        md = self.metadata_by_key.get(key, None)
        if md is not None:
            yield key, md

        for nest_key, nested in (self.nested or {}).items():
            for sub_key, info in nested.find_by_key(key):
                yield f"{nest_key}:{sub_key}", info


@dataclass
class ServerPluginSpec:
    name: str
    #: Python module
    module: Optional[str] = None
    #: Or any executable
    executable: Optional[List[str]] = None
    #: Can be a dataclass or a builtin type
    # result_class: type
    files_to_monitor: Dict[str, str] = field(default_factory=dict)
    results: Optional[PluginResults] = None
    results_json: Any = None
    # Require IOCs to be loaded first before running
    after_iocs: bool = False

    @property
    def script(self) -> str:
        """The script and arguments to run for the plugin."""
        if self.executable:
            return " ".join(f'"{param}"' for param in self.executable)
        elif self.module:
            return f'"{sys.executable}" -m {self.module}'
        raise ValueError("module and executable both unset")

    async def update(self) -> Optional[PluginResults]:
        """Call the plugin and get new information, storing it in results."""
        self.results_json = (
            await util.run_script_with_json_output(self.script)
        ) or {}
        self.files_to_monitor = self.results_json.get("files_to_monitor", {})
        self.results = apischema.deserialize(PluginResults, self.results_json)
        if self.results:
            self.files_to_monitor = self.results.files_to_monitor
        return self.results


@dataclass
class PVGetInfo:
    pv_name: str
    present: bool
    info: List[WhatRecord]

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": """\
{{ pv_name }}:
    In database: {{ present }}
{% for _info in info %}
{% set item_info = render_object(_info, "console") %}
    {{ item_info | indent(4)}}
{% endfor %}
}
""",
    }


@dataclass
class PVGetMatchesResponse:
    patterns: List[str]
    regex: bool
    matches: List[str]


@dataclass
class IocGetMatchesResponse:
    patterns: List[str]
    regex: bool
    matches: List[IocMetadata]


@dataclass
class IocGetDuplicatesResponse:
    patterns: List[str]
    regex: bool
    duplicates: Dict[str, List[str]]


AnyRecordInstance = Union[RecordInstanceSummary, RecordInstance]


@dataclass
class IocGetMatchingRecordsResponse:
    ioc_patterns: List[str]
    record_patterns: List[str]
    regex: bool
    # TODO: ew, redo this
    matches: List[Tuple[IocMetadata, List[AnyRecordInstance]]]


@dataclass
class RecordFieldSummary:
    dtype: str
    name: str
    value: Any

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": """field({{name}}, "{{value}}")""",
        "console-verbose": """\
field({{name}}, "{{value}}")
""",
    }


PVRelationsSummary = Dict[str, List[str]]


@dataclass
class PVRelationshipResponse:
    pv_relations: PVRelations
    script_relations: ScriptPVRelations
    ioc_to_pvs: Dict[str, List[str]]


@dataclass
class PVShortRelationshipResponse:
    # pv_relations: PVRelationsSummary
    script_relations: ScriptPVRelations
    # ioc_to_pvs: Dict[str, List[str]]

    @classmethod
    def from_pv_relations(
        cls,
        pv_relations: PVRelations,
        script_relations: ScriptPVRelations,
        ioc_to_pvs: Dict[str, List[str]],
    ) -> PVRelationshipResponse:
        # summary = {
        #     pv1: list(pv2s)
        #     for pv1, pv2s in pv_relations.items()
        # }
        return cls(
            # pv_relations=summary,
            script_relations=script_relations,
            # ioc_to_pvs=ioc_to_pvs,
        )


@dataclass
class FileInfoRawResponse:
    lines: List[str]
    path: str
    hash: str


@dataclass
class FileInfoResponse:
    script: Optional[IocshScript]
    ioc: Optional[IocMetadata]
    raw: Optional[FileInfoRawResponse] = None



import pytest

from .. import graph
from ..common import LoadContext, RecordField
from ..db import Database


def create_record_def(record_type):
    return f"""
recordtype({record_type}) {{
    field(INP, DBF_INLINK) {{
    }}
    field(OUT, DBF_OUTLINK) {{
    }}
    field(VAL, DBF_STRING) {{
    }}
}}
"""


def create_record(record_type, record_name, fields, filename=None):
    filename = filename or f"{record_name}.db"
    field_source = "\n".join(
        f'field({field_name}, "{field_value}")'
        for field_name, field_value in fields.items()
    )
    db_source = create_record_def(record_type) + f"""
record({record_type}, "{record_name}") {{
    {field_source}
}}
"""
    print(f"---- {filename}:")
    for lineno, line in enumerate(db_source.splitlines(), 1):
        print(lineno, line)

    db = Database.from_string(db_source, filename=filename)
    return db.records[record_name]


def test_simple_graph(dbd: Database):
    database = {
        "record_a": create_record("ai", "record_a", {"INP": "record_b CPP MS", "VAL": "10"}),
        "record_b": create_record("ao", "record_b", {"OUT": "record_c CA", "VAL": "20"}),
    }
    relations = graph.build_database_relations(
        database,
        record_types=dbd.record_types,
    )
    print(database["record_a"])
    assert relations["record_a"]["record_b"] == [
        (
            database["record_a"].fields["INP"],
            database["record_b"].fields["VAL"],
            ["CPP", "MS"],
        ),
    ]
    assert relations["record_b"]["record_a"] == [
        (
            database["record_b"].fields["VAL"],
            database["record_a"].fields["INP"],
            ["CPP", "MS"],
        ),
    ]
    assert relations["record_b"]["record_c"] == [
        (
            database["record_b"].fields["OUT"],
            RecordField(
                dtype='unknown', name='VAL', value='(unknown-record)',
                context=(LoadContext("unknown", 0), ),
            ),
            ["CA"],
        ),
    ]


@pytest.fixture
def dbd():
    return Database.from_string(
        "\n".join((
            create_record_def(rec_type)
            for rec_type in ["ai", "ao"]
        )),
        filename="the.dbd",
    )


def test_combine_relations(dbd: Database):
    database_1 = {
        "record_a": create_record("ai", "record_a", {"INP": "record_b CPP MS", "VAL": "10"}),
        "record_b": create_record("ao", "record_b", {"OUT": "record_c CA", "VAL": "20"}),
    }
    database_2 = {
        "record_c": create_record("ao", "record_c", {"OUT": "record_d.INP", "VAL": "10"}),
        "record_d": create_record("ai", "record_d", {"INP": "", "VAL": "10"}),
        "record_e": create_record("ai", "record_e", {"INP": "record_a CP", "VAL": "10"}),
    }
    relations = graph.build_database_relations(
        database_1,
        record_types=dbd.record_types,
    )
    graph.combine_relations(
        relations,
        database_1,
        graph.build_database_relations(database_2),
        database_2,
        record_types=dbd.record_types,
    )

    assert relations["record_a"]["record_b"] == [
        (
            database_1["record_a"].fields["INP"],
            database_1["record_b"].fields["VAL"],
            ["CPP", "MS"],
        ),
    ]
    assert relations["record_b"]["record_a"] == [
        (
            database_1["record_b"].fields["VAL"],
            database_1["record_a"].fields["INP"],
            ["CPP", "MS"],
        ),
    ]
    assert relations["record_b"]["record_c"] == [
        (
            database_1["record_b"].fields["OUT"],
            # RecordField(
            #     dtype='unknown', name='VAL', value='(unknown-record)',
            #     context=(LoadContext("unknown", 0), ),
            # ),
            database_2["record_c"].fields["VAL"],
            ["CA"],
        ),
    ]
    assert relations["record_c"]["record_d"] == [
        (
            database_2["record_c"].fields["OUT"],
            database_2["record_d"].fields["INP"],
            [],
        ),
    ]
    assert relations["record_d"]["record_c"] == [
        (
            database_2["record_d"].fields["INP"],
            database_2["record_c"].fields["OUT"],
            [],
        ),
    ]
    assert relations["record_e"]["record_a"] == [
        (
            database_2["record_e"].fields["INP"],
            database_1["record_a"].fields["VAL"],
            ["CP"],
        ),
    ]


def test_with_unset_field(dbd: Database):
    database_1 = {
        "record_a": create_record("ai", "record_a", {"INP": "record_b CPP MS", "VAL": "10"}),
        "record_b": create_record("ao", "record_b", {"OUT": "record_c CA"}),
    }
    relations = graph.build_database_relations(database_1, record_types=dbd.record_types)
    assert relations["record_a"]["record_b"] == [
        (
            database_1["record_a"].fields["INP"],
            # database_1["record_b"].fields["VAL"],  # <-- not defined, but in dbd
            RecordField(
                dtype="DBF_STRING",
                name="VAL",
                value="",
                context=(LoadContext("the.dbd", 17),),
            ),
            ["CPP", "MS"],
        ),
    ]


def test_with_invalid_field(dbd: Database):
    database_1 = {
        "record_a": create_record("ai", "record_a", {"INP": "record_b.INVAL CPP MS", "VAL": "10"}),
        "record_b": create_record("ao", "record_b", {"OUT": "record_c CA"}),
    }
    relations = graph.build_database_relations(database_1, record_types=dbd.record_types)
    assert relations["record_a"]["record_b"] == [
        (
            database_1["record_a"].fields["INP"],
            RecordField(
                dtype="invalid",
                name="INVAL",
                value="(invalid-field)",
                context=(LoadContext("unknown", 0),),
            ),
            ["CPP", "MS"],
        ),
    ]


def test_with_weird_record_type(dbd: Database):
    database_1 = {
        "record_a": create_record("ai", "record_a", {"INP": "record_b.ABC CPP MS", "VAL": "10"}),
        "record_b": create_record("aq", "record_b", {"OUT": "record_c CA"}),
    }
    relations = graph.build_database_relations(database_1, record_types=dbd.record_types)
    assert relations["record_a"]["record_b"] == [
        (
            database_1["record_a"].fields["INP"],
            RecordField(
                dtype="invalid",
                name="ABC",
                value="(invalid-record-type)",
                context=(LoadContext("unknown", 0),),
            ),
            ["CPP", "MS"],
        ),
    ]


def test_with_alias(dbd: Database):
    database_1 = {
        "record_a": create_record("ai", "record_a", {"INP": "foobar CPP MS", "VAL": "10"}),
        "record_b": create_record("ao", "record_b", {"VAL": "5"}),
    }
    relations = graph.build_database_relations(
        database_1, record_types=dbd.record_types,
        aliases={"foobar": "record_b"},
    )
    assert relations["record_a"]["record_b"] == [
        (
            database_1["record_a"].fields["INP"],
            database_1["record_b"].fields["VAL"],
            ["CPP", "MS"],
        ),
    ]


def test_combine_with_alias(dbd: Database):
    database_1 = {
        "record_a": create_record("ai", "record_a", {"INP": "record_b CPP MS", "VAL": "10"}),
        "record_b": create_record("ao", "record_b", {"OUT": "alias_c CA", "VAL": "20"}),
    }
    database_2 = {
        "record_c": create_record("ao", "record_c", {"OUT": "alias_d.INP", "VAL": "10"}),
        # Refer to record_a through
        "record_d": create_record("ai", "record_d", {"INP": "record_a", "VAL": "10"}),
        # ... but _also_ refer to it as alias_a; when combining the relations this makes
        # it more fun
        "record_e": create_record("ai", "record_e", {"INP": "alias_a CP", "VAL": "10"}),
    }
    relations = graph.build_database_relations(
        database_1,
        aliases={
            "alias_a": "record_a",
            "alias_b": "record_b",
        },
        record_types=dbd.record_types,
    )

    relations_2 = graph.build_database_relations(
        database_2,
        aliases={
            "alias_c": "record_c",
            "alias_d": "record_d",
            "alias_e": "record_e",
        },
        record_types=dbd.record_types,
    )

    graph.combine_relations(
        relations,
        database_1,
        relations_2,
        database_2,
        aliases={
            "alias_a": "record_a",
            "alias_b": "record_b",
            "alias_c": "record_c",
            "alias_d": "record_d",
            "alias_e": "record_e",
        },
        record_types=dbd.record_types,
    )

    assert relations["record_a"]["record_b"] == [
        (
            database_1["record_a"].fields["INP"],
            database_1["record_b"].fields["VAL"],
            ["CPP", "MS"],
        ),
    ]
    assert relations["record_b"]["record_a"] == [
        (
            database_1["record_b"].fields["VAL"],
            database_1["record_a"].fields["INP"],
            ["CPP", "MS"],
        ),
    ]
    assert relations["record_b"]["record_c"] == [
        (
            database_1["record_b"].fields["OUT"],
            # RecordField(
            #     dtype='unknown', name='VAL', value='(unknown-record)',
            #     context=(LoadContext("unknown", 0), ),
            # ),
            database_2["record_c"].fields["VAL"],
            ["CA"],
        ),
    ]
    assert relations["record_c"]["record_d"] == [
        (
            database_2["record_c"].fields["OUT"],
            database_2["record_d"].fields["INP"],
            [],
        ),
    ]
    assert relations["record_d"]["record_c"] == [
        (
            database_2["record_d"].fields["INP"],
            database_2["record_c"].fields["OUT"],
            [],
        ),
    ]
    assert relations["record_d"]["record_a"] == [
        (
            database_2["record_d"].fields["INP"],
            database_1["record_a"].fields["VAL"],
            [],
        ),
    ]
    assert relations["record_e"]["record_a"] == [
        (
            database_2["record_e"].fields["INP"],
            database_1["record_a"].fields["VAL"],
            ["CP"],
        ),
    ]
    assert relations["record_a"]["record_e"] == [
        (
            database_1["record_a"].fields["VAL"],
            database_2["record_e"].fields["INP"],
            ["CP"],
        ),
    ]


def test_list_value_graph(dbd: Database):
    database_1 = {
        "record_a": create_record(
            "ai",
            "record_a",
            {
                # Note: field(INP, ["this is a long string array value"])
                "INP": ("this is a long string array value",),
                "VAL": "10",
            },
        ),
    }

    # Ensure this doesn't raise
    graph.build_database_relations(
        database_1,
        record_types=dbd.record_types,
    )



import contextlib
import functools
import sys


@contextlib.contextmanager
def suppress_output():
    class OutputBuffer:
        def __init__(self):
            self.buffer = []

        def write(self, buf):
            self.buffer.append(buf)

        def flush(self):
            ...

    replacement_stderr = OutputBuffer()
    sys.stderr = replacement_stderr

    replacement_stdout = OutputBuffer()
    sys.stdout = replacement_stdout

    try:
        yield replacement_stdout, replacement_stderr
    finally:
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__


def suppress_output_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with suppress_output() as (buffered_stdout, buffered_stderr):
            results = func(*args, **kwargs)
            results.execution_info["stdout"] = "\n".join(buffered_stdout.buffer)
            results.execution_info["stderr"] = "\n".join(buffered_stderr.buffer)
            return results
    return wrapper



from .access_security import AccessSecurityConfig
from .common import FileFormat
from .db import Database
from .dbtemplate import TemplateSubstitution
from .gateway import GatewayConfig
from .gateway import PVList as GatewayPVList
from .iocsh import parse_iocsh_line
from .macro import MacroContext
from .parse import parse
from .plugins.epicsarch import LclsEpicsArchFile
from .snl import SequencerProgram
from .streamdevice import StreamProtocol
from .version import __version__  # noqa: F401

__all__ = [
    "AccessSecurityConfig",
    "Database",
    "FileFormat",
    "GatewayConfig",
    "GatewayPVList",
    "LclsEpicsArchFile",
    "MacroContext",
    "SequencerProgram",
    "StreamProtocol",
    "TemplateSubstitution",
    "parse_iocsh_line",
    "parse",
]



from __future__ import annotations

import enum
import functools
import inspect
import json
import logging
import pathlib
import textwrap
import typing
from contextlib import contextmanager
from dataclasses import dataclass, field
from time import perf_counter
from typing import (Any, Callable, ClassVar, Dict, Generator, List, Optional,
                    Sequence, Tuple, Union)

import apischema
import epicsmacrolib
import lark

if typing.TYPE_CHECKING:
    from .shell import ShellState

from . import settings, util

logger = logging.getLogger(__name__)


class FileFormat(str, enum.Enum):
    """
    File type enum that covers all formats supported by whatrecord.
    """
    iocsh = 'iocsh'
    database = 'database'
    database_definition = 'database_definition'
    substitution = 'substitution'
    gateway_pvlist = 'gateway_pvlist'
    access_security = 'access_security'
    stream_protocol = 'stream_protocol'
    state_notation = 'state_notation'
    makefile = 'makefile'

    @classmethod
    def from_extension(cls, extension: str) -> FileFormat:
        """Get a file format based on a file extension."""
        return {
            "cmd": FileFormat.iocsh,
            "db": FileFormat.database,
            "dbd": FileFormat.database_definition,
            "template": FileFormat.database,
            "substitutions": FileFormat.substitution,
            "pvlist": FileFormat.gateway_pvlist,
            "acf": FileFormat.access_security,
            "proto": FileFormat.stream_protocol,
            "st": FileFormat.state_notation,
        }[extension.lower()]

    @classmethod
    def from_filename(cls, filename: AnyPath) -> FileFormat:
        """Get a file format based on a full filename."""
        path = pathlib.Path(filename)
        extension = path.suffix.lstrip(".")
        if not extension and path.name.startswith("Makefile"):
            return FileFormat.makefile

        try:
            return FileFormat.from_extension(extension)
        except KeyError:
            raise ValueError(
                f"Could not determine file type from extension: {extension}"
            ) from None


@dataclass(frozen=True)
class LoadContext:
    """File and line context information."""
    #: The filename (or other identifier) to identify the context.
    name: str
    #: The line number of ``name`` this applies to.
    line: int

    def __repr__(self):
        return f"{self.name}:{self.line}"

    @apischema.serializer
    @property
    def as_tuple(self) -> Sequence[Union[str, int]]:
        return [self.name, self.line]


@apischema.deserializer
def _load_context_from_tuple(items: Sequence[Union[str, int]]) -> LoadContext:
    """
    A deserializer that takes in e.g., ["file", line] and turns it into a LoadContext.
    """
    return LoadContext(*items)


# FullLoadContext = Tuple[LoadContext, ...]
FullLoadContext = Sequence[LoadContext]
IocInfoDict = Dict[str, Union[str, Dict[str, str], List[str]]]
AnyPath = Union[str, pathlib.Path]


@dataclass(repr=False)
class MutableLoadContext:
    """
    A mutable (i.e., changeable) version of :class:`LoadContext`.
    """
    name: str
    line: int

    def __repr__(self):
        return f"{self.name}:{self.line}"

    def to_load_context(self) -> LoadContext:
        """Convert this to an immutable LoadContext."""
        return LoadContext(self.name, self.line)


@dataclass
class IocshCmdArgs:
    """iocshCmd(...) arguments."""
    context: FullLoadContext
    command: str


@apischema.fields.with_fields_set
@dataclass
class IocshResult:
    context: FullLoadContext
    line: str
    outputs: List[str] = field(default_factory=list)
    argv: Optional[List[str]] = None
    error: Optional[str] = None
    redirects: List[epicsmacrolib.IocshRedirect] = field(default_factory=list)
    # TODO: normalize this
    # result: Optional[Union[str, Dict[str, str], IocshCmdArgs, ShortLinterResults]]
    result: Any = None

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": """
{%- for ctx in context -%}
    {{ctx.line}}:
{%- endfor %} {{ line }}
{% if error %}
    ** ERROR on line {% if context %}{{ context[0].line }}{% endif %} **
    {{ error | indent(_fmt.indent) }}
{% endif %}
{% if outputs != [line] %}
{% for output in outputs %}
    -SH> {{ output | indent(6) }}
{% endfor %}
{% endif %}
{% for redirect in redirects %}
    -> Redirect: {{ redirect }}
{%- endfor %}
""".strip(),
    }


@dataclass
class IocshScript:
    """An IOC Shell script (i.e., st.cmd) that was loaded and interpreted."""
    #: The path to the script.
    path: str
    #: Interpreted lines of the script.
    lines: List[IocshResult]

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            {%- for line in lines %}
            {% if line.context | length == 1 %}
            {{ line.line }}
            {% endif %}
            {%- endfor -%}
            """.rstrip(),
        ),
        "console": textwrap.dedent(
            """\
            {%- for line in lines %}
            {{ render_object(line, "console") }}
            {%- endfor %}
            """.rstrip(),
        ),
        "console-verbose": textwrap.dedent(
            """\
            {%- for line in lines -%}
            {{ render_object(line, "console-verbose") }}
            {%- endfor %}
            """.rstrip(),
        )
    }

    @classmethod
    def from_metadata(cls, md: IocMetadata, sh: ShellState) -> IocshScript:
        """
        Create an IocshScript provided IocMetadata and ShellState.

        Parameters
        ----------
        md : IocMetadata
            Metadata identifying the IOC.
        sh : ShellState
            The state of the shell interpreter after interpreting the IOC
            startup script.

        Returns
        -------
        IocshScript
        """
        if md.looks_like_sh:
            if md.base_version == settings.DEFAULT_BASE_VERSION:
                md.base_version = "unknown"
            return cls.from_general_file(md.script)

        return cls(
            path=str(md.script),
            lines=tuple(
                sh.interpret_shell_script(
                    md.script
                )
            ),
        )

    @classmethod
    def from_interpreted_script(
        cls,
        filename: Union[pathlib.Path, str],
        contents: str,
        sh: ShellState
    ) -> IocshScript:
        """
        Create an IocshScript provided a filename, its contents, and ShellState.

        Parameters
        ----------
        filename : str
            The script filename.
        contents : str
            The decoded string contents of the file.
        sh : ShellState
            The state of the shell interpreter after interpreting the IOC
            startup script.

        Returns
        -------
        IocshScript
        """
        return cls(
            path=str(filename),
            lines=tuple(
                sh.interpret_shell_script_text(
                    contents.splitlines(),
                    name=str(filename)
                )
            ),
        )

    @classmethod
    def from_general_file(cls, filename: Union[pathlib.Path, str]) -> IocshScript:
        """
        Create an IocshScript from any given file, even non-startup scripts.

        This can be used to transfer database files over to the frontend, for
        example.

        Parameters
        ----------
        filename : Union[pathlib.Path, str]
            The filename.

        Returns
        -------
        """
        # For use when shoehorning in a file that's not _really_ an IOC script
        # TODO: instead rework the api
        with open(filename, "rt") as fp:
            lines = fp.read().splitlines()

        return cls(
            path=str(filename),
            lines=tuple(
                IocshResult(
                    line=line,
                    context=(LoadContext(str(filename), lineno), )
                )
                for lineno, line in enumerate(lines, 1)
            ),
        )


@dataclass
class IocshArgument:
    """A single argument in a shell command."""
    #: The name of the argument.
    name: str
    #: Its type, according to EPICS.
    type: str


@dataclass
class IocshCommand:
    """
    A registered shell command.

    Largely used as part of the gdb binary information tool.
    """
    #: The name of the command.
    name: str
    #: Arguments the user can pass to the command.
    args: List[IocshArgument] = field(default_factory=list)
    #: Usage information.
    usage: Optional[str] = None
    #: Where this command information came from.
    context: Optional[FullLoadContext] = None


@dataclass
class IocshVariable:
    """
    A registered shell variable.

    Largely used as part of the gdb binary information tool.
    """
    #: The name of the variable.
    name: str
    #: The last value of the variable.
    value: Optional[str] = None
    #: The type of the variable, according to EPICS.
    type: Optional[str] = None


@dataclass
class GdbBinaryInfo:
    """
    GDB-derived binary information.

    This is populated by deserializing the output of
    ``whatrecord.plugins.gdb_binary_info``, which is interpreted by GDB itself.
    """
    #: Commands that are available to the user in an EPICS IOC shell.
    commands: Dict[str, IocshCommand]
    #: The base version detected in the binary.
    base_version: Optional[str]
    #: Variables available to be set in the shell.
    variables: Dict[str, IocshVariable]
    #: GDB plugin error information.
    error: Optional[str]


def get_grammar_version_by_base_version(base_version: str) -> int:
    """
    Database grammar version to use, provided the epics-base version.

    Returns
    -------
    base_version : str
        The epics-base version number.

    Returns
    -------
    int
        If R3.15 or under, ``3``, otherwise ``4``.
    """
    base_version = base_version.lstrip("vRr")
    major_minor = tuple(
        int(v) for v in base_version.split(".")[:2]
    )
    return 3 if major_minor < (3, 16) else 4


@dataclass
class IocMetadata:
    """
    Metadata identifying an IOC.
    """
    #: The name of the IOC.
    name: str = "unset"
    #: The IOC startup script.
    script: pathlib.Path = field(default_factory=pathlib.Path)
    #: The startup directory to use when interpreting the script.
    startup_directory: pathlib.Path = field(default_factory=pathlib.Path)
    #: The host that the IOC will be run on, if available.
    host: Optional[str] = None
    #: The port on the host that the IOC will be available on, if available.
    port: Optional[int] = None
    #: The path to the IOC binary.
    binary: Optional[str] = None
    #: The EPICS-base version.
    base_version: str = settings.DEFAULT_BASE_VERSION
    #: User-specified metadata about the IOC.
    metadata: Dict[str, Any] = field(default_factory=dict)
    #: Macros to include when loading the IOC.
    macros: Dict[str, str] = field(default_factory=dict)
    #: Stand-in directories to use when interpreting a script outside of its
    #: normal environment.
    standin_directories: Dict[str, str] = field(default_factory=dict)
    #: Commands available to be used in the IOC shell.
    commands: Dict[str, IocshCommand] = field(default_factory=dict)
    #: Variables that are defined for usage in the shell.
    variables: Dict[str, IocshVariable] = field(default_factory=dict)
    #: Files that were loaded with their hash as a value.
    loaded_files: Dict[str, str] = field(default_factory=dict)
    #: Whether the IOC was successfully loaded or not.
    load_success: bool = True

    def update(self, other: IocMetadata, merge: bool = False):
        """
        Update the metadata with a new set from an IOC Loader.

        Parameters
        ----------
        other : IocMetadata
            The other IOC metadata to update this instance with.

        merge : bool, optional
            Merge in dictionary information, or overwrite it.  Defaults to
            'overwrite' (merge=False).
        """
        self.name = other.name
        self.script = other.script
        self.startup_directory = other.startup_directory
        self.host = other.host or self.host
        self.port = other.port or self.port
        self.binary = other.binary or self.binary
        self.base_version = (
            other.base_version
            if other.base_version != settings.DEFAULT_BASE_VERSION
            else self.base_version
        )

        if merge:
            self.metadata.update(other.metadata)
            self.macros.update(other.macros)
            self.standin_directories.update(other.standin_directories)
            self.commands.update(other.commands)
            self.variables.update(other.variables)
        else:
            self.metadata = dict(other.metadata)
            self.macros = dict(other.macros)
            self.standin_directories = dict(other.standin_directories)
            self.commands = dict(other.commands)
            self.variables = dict(other.variables)

    @property
    def looks_like_sh(self) -> bool:
        """Is the script likely sh/bash/etc?"""
        return self.binary and (
            "bin/sh" in self.binary or
            "bin/bash" in self.binary or
            "env bash" in self.binary or
            "bin/tcsh" in self.binary or
            "/python" in self.binary
        )

    @property
    def _cache_key(self) -> str:
        """Cache key for storing this IOC information in a file."""
        key = "".join(
            str(v)
            for v in (
                self.name,
                str(self.script.resolve()),
                str(self.startup_directory.resolve()),
                str(self.host),
                str(self.port),
            )
        )
        hash = util.get_bytes_sha256(bytes(key, "utf-8"))
        return f"{self.name}.{hash}"

    @property
    def cache_filename(self) -> pathlib.Path:
        """The cache filename for this metadata."""
        metadata_fn = f"{self._cache_key}.IocMetadata.json"
        return pathlib.Path(settings.CACHE_PATH) / metadata_fn

    @property
    def ioc_cache_filename(self) -> pathlib.Path:
        """The cache filename for the entire interpreted IOC."""
        metadata_fn = f"{self._cache_key}.LoadedIoc.json"
        return pathlib.Path(settings.CACHE_PATH) / metadata_fn

    def from_cache(self) -> Optional[IocMetadata]:
        """Load metadata from the cache, if available."""
        if not settings.CACHE_PATH:
            return

        try:
            with open(self.cache_filename, "rb") as fp:
                return apischema.deserialize(type(self), json.load(fp))
        except FileNotFoundError:
            ...
        except json.decoder.JSONDecodeError:
            # Truncated output file, perhaps
            ...

    def save_to_cache(self) -> bool:
        """Save this metadata to the cache, if enabled."""
        if not settings.CACHE_PATH:
            return False

        with open(self.cache_filename, "wt") as fp:
            json.dump(apischema.serialize(self), fp=fp)
        return True

    def is_up_to_date(self) -> bool:
        """Is this IOC up-to-date with what is on disk?"""
        if not self.loaded_files:
            return False
        return util.check_files_up_to_date(self.loaded_files)

    def add_loaded_file(
        self,
        filename: Union[pathlib.Path, str],
        update: bool = False,
    ) -> bool:
        """
        Add filename to the loaded file dictionary, optionally updating an
        existing hash.

        Parameters
        ----------
        filename : Union[pathlib.Path, str]
            The filename.
        update : bool, optional
            Update a hash if the filename is already in the loaded_files
            dictionary.

        Returns
        -------
        bool
            Set if a hash was updated in the loaded_files dictionary.
        """
        filename = pathlib.Path(self.startup_directory) / filename
        if str(filename) not in self.loaded_files or update:
            self.loaded_files[str(filename)] = util.get_file_sha256(filename)
            return True
        return False

    async def get_binary_information(self) -> Optional[GdbBinaryInfo]:
        """
        Get binary information using the GDB plugin.

        Requires that gdb be available, along with the ``.binary`` be set
        to a valid file.
        """
        if not self.binary or not pathlib.Path(self.binary).exists():
            return

        try:
            info = await util.run_gdb(
                "gdb_binary_info",
                self.binary,
                cls=GdbBinaryInfo,
            )
        except apischema.ValidationError:
            logger.error(
                "Failed to get gdb information for %s (%s)",
                self.name, self.binary,
                exc_info=True,
            )
            return

        if info.error:
            logger.error(
                "Failed to get gdb information for %s (%s): %s",
                self.name, self.binary, info.error
            )
            return

        self.base_version = info.base_version or self.base_version
        self.commands.update(info.commands)
        self.variables.update(info.variables)

        for command in self.commands.values():
            for context in command.context or []:
                try:
                    self.add_loaded_file(context.name)
                except FileNotFoundError:
                    logger.debug(
                        "GDB source file does not exist for command %s (%s)",
                        command.name, context
                    )
        return info

    @property
    def database_version_spec(self) -> int:
        """
        Load databases with this specification.

        Returns
        -------
        version_spec : int
            If R3.15 or under, ``3``, otherwise ``4``.
        """
        # TODO: better version parsing

        for base_version in [self.base_version, settings.DEFAULT_BASE_VERSION]:
            try:
                return get_grammar_version_by_base_version(base_version)
            except Exception:
                ...

        return 3

    @classmethod
    def from_file(
        cls,
        filename: Union[pathlib.Path, str],
        *,
        name: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        startup_directory: Optional[pathlib.Path] = None,
        macros: Optional[Dict[str, str]] = None,
        standin_directories: Optional[Dict[str, str]] = None,
        binary: Optional[str] = None,
        base_version: Optional[str] = settings.DEFAULT_BASE_VERSION,
        **metadata
    ) -> IocMetadata:
        """
        Given at minimum a filename, guess the rest of the metadata.

        Parameters
        ----------
        filename : pathlib.Path or str
            The IOC startup script filename.
        name : str, optional
            The name of the IOC.  Defaults to the startup script's parent
            directory name.
        host : str, optional
            The host on which the IOC runs.
        port : int, optional
            The port on which the IOC will be accessible from the host.
        startup_directory : pathlib.Path, optional
            The directory in which the script should be started.  Defaults
            to the parent directory of ``filename``.
        macros : Dict[str, str], optional
            Macros to use when interpreting the startup script.
        standin_directories : Dict[str, str], optional
            Stand-in directories to use when interpreting a script outside of
            its normal environment.
        binary : str, optional
            The binary used to run the startup script normally. That is, an
            EPICS IOC binary.  If not specified, this will be determined
            based on the startup script hashbang (if available).
        base_version : str, optional
            The epics-base version to use when interpreting the startup script.
            This may be overridden automatically when interpreting the startup
            script.
        **metadata :
            User-specified metadata.

        Returns
        -------
        IocMetadata
        """
        filename = pathlib.Path(filename).expanduser().resolve()
        name = name or filename.parts[-2]  # iocBoot/((ioc-something))/st.cmd
        if "/" in name:
            name = name.replace("/", "")
            if not name:
                # Sorry, we need something unique here (better ideas welcome)
                name = util.get_bytes_sha256(bytes(str(filename), "utf-8"))[:10]

        return cls(
            name=name,
            host=host,
            port=port,
            script=filename,
            startup_directory=startup_directory or filename.parent,
            metadata=metadata,
            macros=macros or {},
            standin_directories=standin_directories or {},
            binary=binary or util.find_binary_from_hashbang(filename),
            base_version=base_version,
        )

    #: Back-compat: from_filename is deprecated
    from_filename = from_file

    @classmethod
    def from_dict(
        cls, iocdict: IocInfoDict, macros: Optional[Dict[str, str]] = None
    ) -> IocMetadata:
        """
        Create an IocMetadata instance.

        Pick apart a given ``iocdict`` dictionary, relegating extra info to
        ``.metadata``.

        Parameters
        ----------
        iocdict : dict
            IOC information dictionary.
        macros : dict, optional
            Additional macros to use at startup.
        """
        ioc = dict(iocdict)
        name = ioc.pop("name")
        host = ioc.pop("host", None)
        port = ioc.pop("port", None)

        script = ioc.pop("script")
        script = pathlib.Path(str(script)).expanduser().resolve()
        binary = ioc.pop("binary", None)
        base_version = ioc.pop("base_version", None)
        return cls(
            name=name,
            script=script,
            startup_directory=script.parent,
            host=host,
            port=port,
            metadata=ioc,
            macros=macros or {},
            binary=binary or util.find_binary_from_hashbang(script),
            base_version=base_version or settings.DEFAULT_BASE_VERSION,
        )


@dataclass
class LinterMessage:
    name: str
    context: FullLoadContext
    message: str


@dataclass
class LinterWarning(LinterMessage):
    ...


@dataclass
class LinterError(LinterMessage):
    ...


PVAJsonField = Dict[str, str]


@dataclass
class RecordField:
    """
    A field with a value, as part of a EPICS V3 RecordInstance.
    """
    dtype: str
    name: str
    value: Any
    context: FullLoadContext

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": """field({{name}}, "{{value}}")""",
        "console-verbose": textwrap.dedent(
            """\
            field({{name}}, "{{value}}")  # {{dtype}}
            {%- if context %}; {{context[-1]}}{% endif %}\
            """.rstrip(),
        ),
        "file": """
            field({{name}}, {{ _json_dump(value) }})
        """.strip(),
    }

    def update_from_record_type(
        self,
        record_type: RecordType
    ):
        """Update field information given dbd-provided information."""
        record_type_field = record_type.fields.get(self.name, None)
        if record_type_field is not None:
            self.dtype = record_type_field.type

    def update_unknowns(
        self,
        other: RecordField,
        *,
        unknown_values: Optional[Sequence[str]] = None,
    ):
        """
        If this RecordField has some missing information ("unknown"), fill
        it in with information from the other field.
        """
        unknown_values = unknown_values or {"unknown", "", "(unknown-record)"}
        if other.dtype not in unknown_values and self.dtype in unknown_values:
            self.dtype = other.dtype
        if other.value not in unknown_values and self.value in unknown_values:
            self.value = other.value
        if len(other.context) and len(self.context) == 1:
            ctx, = self.context
            if ctx.name in unknown_values:
                # Even if the other context is unknown, let's take it anyway:
                self.context = other.context


# field1, field2, options (CA, CP, CPP, etc.)
FieldRelation = Tuple[RecordField, RecordField, List[str]]

PVRelations = Dict[
    str, Dict[str, List[FieldRelation]]
]


ScriptPVRelations = Dict[
    str, Dict[str, List[str]]
]


def get_link_information(link_str: str) -> Tuple[str, List[str]]:
    """Get link information from a DBF_{IN,OUT,FWD}LINK value."""
    if isinstance(link_str, (dict, list)):
        raise ValueError("JSON and PVAccess links are a TODO (sorry!)")
    if not isinstance(link_str, str):
        raise ValueError(
            f"Unexpected and supported type for get_link_information: "
            f"{type(link_str)}"
        )

    if " " in link_str:
        # strip off PP/MS/etc (TODO might be useful later)
        link_str, additional_info = link_str.split(" ", 1)
    else:
        additional_info = ""

    if link_str.startswith("@"):
        # TODO asyn/device links
        raise ValueError("asyn link")
    if not link_str:
        raise ValueError("empty link")

    if link_str.isnumeric():
        # 0 or 1 usually and not a string
        raise ValueError("integral link")

    try:
        float(link_str)
    except Exception:
        # Good, we don't want a float
        ...
    else:
        raise ValueError("float link")

    link_details = additional_info.split()
    return link_str, link_details


LINK_TYPES = {"DBF_INLINK", "DBF_OUTLINK", "DBF_FWDLINK"}


@dataclass
class RecordInstanceSummary:
    """An abbreviated form of :class:`RecordInstance`."""

    context: FullLoadContext
    name: str
    record_type: str
    # fields: Dict[str, RecordField]
    info: Dict[StringWithContext, Any] = field(default_factory=dict)
    # metadata: Dict[str, Any] = field(default_factory=dict)
    aliases: List[str] = field(default_factory=list)
    # is_grecord: bool = False
    is_pva: bool = False
    owner: str = ""

    @classmethod
    def from_record_instance(self, instance: RecordInstance) -> RecordInstanceSummary:
        return RecordInstanceSummary(
            context=instance.context,
            name=instance.name,
            record_type=instance.record_type,
            info=instance.info,
            aliases=instance.aliases,
            is_pva=instance.is_pva,
            owner=instance.owner,
        )


class UnquotedString(str):
    """
    An unquoted string token found when loading a database file.
    May be a linter warning.
    """
    ...


class StringWithContext(str):
    """A string with LoadContext."""
    __slots__ = ("context", )
    context: Optional[FullLoadContext]

    def __new__(cls, value, context: Optional[FullLoadContext] = None):
        self = super().__new__(cls, value)
        self.context = context
        return self


@dataclass
class PVAFieldReference:
    """
    Part of a PVAccess group which references another record.
    """
    context: FullLoadContext
    #: THe name of this field.
    name: str = ""
    #: The referenced record name.
    record_name: str = ""
    #: The referenced record's field.
    field_name: str = ""
    #: Metadata from the field definition.
    metadata: Dict[str, str] = field(default_factory=dict)

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": """\
PVAFieldReference: {{ record_name }}.{{ field_name }}
                 - {{ metadata }}
""",
        "file": "",
    }


AnyField = Union[RecordField, PVAFieldReference]


@dataclass
class DatabaseMenu:
    """An enumeration (menu) from a dbd file."""
    context: FullLoadContext
    #: The name of the menu.
    name: str
    #: Possible options for the menu - user-facing string identifier to
    #: internal identifier.
    choices: Dict[str, str]

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            menu({{ name }}) {
            {% for name, choice in choices.items() %}
            {{ _indent }}choice({{ name }}, "{{ choice }}")
            {% endfor %}
            }
            """.rstrip(),
        ),
    }


@dataclass
class DatabaseDevice:
    """A per-record-type device definition, part of a dbd file."""
    record_type: str
    link_type: str
    dset_name: str
    choice_string: str

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            device({{ record_type }}, {{ link_type }}, {{ dset_name }}, "{{ choice_string }}")
            """.rstrip(),
        ),
    }


@apischema.fields.with_fields_set
@dataclass
class RecordTypeField:
    """
    An EPICS V3 RecordType field definition, part of a database definition (dbd).
    """
    context: FullLoadContext
    #: Record type name.
    name: str
    #: The data type (e.g., DBF_STRING)
    type: str
    #: Access security level (ASL0, ASL1, or None)
    asl: Optional[str] = None
    #: Initial value for the field.
    initial: Optional[str] = None
    #: Grouping for the field in user interfaces.
    promptgroup: Optional[str] = None
    #: User interface information about the field.
    prompt: Optional[str] = None
    #: "Special" settings such as SPC_NOMOD (no write access)
    #: SPC_NOMOD      Field must not be modified
    #: SPC_DBADDR     db_name_to_addr must call cvt_dbaddr
    #: SPC_SCAN       scan related field is being changed
    #: SPC_ALARMACK   Special Alarm Acknowledgement
    #: SPC_AS         Access Security
    #: SPC_ATTRIBUTE  psuedo field, i.e. attribute field
    #:                useful when record support must be notified of a field
    #:                changing value
    #: SPC_MOD        used by all records that support a reset field
    #: SPC_RESET      The res field is being modified
    #:
    #: Specific to conversion:
    #: SPC_LINCONV  A linear conversion field is being changed
    #:
    #: Specific to calculation records
    #: SPC_CALC     The CALC field is being changed
    special: Optional[str] = None
    #: Process passive.
    pp: Optional[str] = None
    #: Process passive. TURE/FALSE.
    interest: Optional[str] = None
    #: Base: e.g., HEX
    base: Optional[str] = None
    #: Number of elements.
    size: Optional[str] = None
    #: Extra debug information.
    extra: Optional[str] = None
    #: Enum (menu) options for the field.
    menu: Optional[str] = None
    #: Property- YES/NO
    prop: Optional[str] = None
    # -> bundle the remainder in "body", even if unrecognized
    body: Dict[str, str] = field(default_factory=dict)

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            field({{ name }}, {{ type }}) {
            {{ obj._get_file_repr() | indent(_fmt.indent) }}
            }
            """.rstrip(),
        ),
    }

    def _get_file_repr(self) -> str:
        # This is too slow to let jinja take care of....
        def get_value_repr(value: str):
            if isinstance(value, UnquotedString):
                return str(value)
            return f'"{value}"'

        return "\n".join(
            f"{attr}({get_value_repr(value)})"
            for attr, value in self.get_all_set_entries().items()
        )

    def get_all_set_entries(self) -> Dict[str, str]:
        """
        Get all field entries that have been set in the database definition.

        Returns
        -------
        dict
            e.g., {"initial": "5"}
        """
        entries = {}
        for attr in (
            # Order defined in dbStaticLib.c
            "prompt",
            "initial",
            "promptgroup",
            "special",
            "extra",
            "menu",
            "size",
            "pp",
            "prop",
            "base",
            "interest",
            "asl",
        ):
            value = getattr(self, attr)
            if value is not None:
                entries[attr] = value
        entries.update(self.body)
        return entries


@dataclass
class RecordType:
    """
    An EPICS V3 record type definition, part of a database definition (dbd).
    """
    context: FullLoadContext
    name: str
    cdefs: List[str] = field(default_factory=list)
    fields: Dict[str, RecordTypeField] = field(default_factory=dict)
    devices: List[DatabaseDevice] = field(default_factory=list)

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            recordtype({{ name }}) {
            {% for cdef in cdefs %}
            {{ _indent }}%{{ cdef }}
            {% endfor %}
            {% for field_name, field in fields.items() %}
            {{ _indent }}field({{ field.name }}, {{ field.type }}) {
            {{ _indent }}{{ _indent }}{{ field._get_file_repr() | indent(_fmt.indent * 2) }}
            {{ _indent }}}
            {% endfor %}
            }
            """.rstrip(),
        ),
    }

    def get_links_for_record(
        self, record: RecordInstance
    ) -> Generator[Tuple[RecordField, str, List[str]], None, None]:
        """
        Get all links - in, out, and forward links - for the given record.

        Parameters
        ----------
        record : RecordInstance
            Additional information, if the database definition wasn't loaded
            with this instance.

        Yields
        ------
        field : RecordField
        link_text: str
        link_info: str
        """
        if record.record_type != self.name:
            raise ValueError("Record types do not match")

        for field_type_info in self.get_fields_of_type(*LINK_TYPES):
            field_instance = record.fields.get(field_type_info.name, None)
            if field_instance is None:
                continue
            elif isinstance(field_instance, PVAFieldReference):
                continue

            try:
                link, info = get_link_information(field_instance.value)
            except ValueError:
                continue
            yield field_instance, link, info

    def get_fields_of_type(self, *types: str) -> Generator[RecordTypeField, None, None]:
        """Get all fields of the matching type(s)."""
        for fld in self.fields.values():
            if fld.type in types:
                yield fld

    def get_link_fields(self) -> Generator[RecordTypeField, None, None]:
        """
        Get all link fields - in, out, and forward links.

        Yields
        ------
        field : RecordTypeField
        """
        yield from self.get_fields_of_type(*LINK_TYPES)


@dataclass
class RecordInstance:
    """
    An instance of a RecordType, loaded from a db file.
    """
    context: FullLoadContext
    #: The name of the record.
    name: str
    #: The record type name.
    record_type: str
    #: Whether or not there is corresponding information in the dbd file.
    has_dbd_info: bool = False
    #: Field information - field name to details.  May be either a V3
    #: RecordField or a PVAFieldReference.
    fields: Dict[str, AnyField] = field(default_factory=dict)
    #:
    info: Dict[StringWithContext, Any] = field(default_factory=dict)
    metadata: Dict[StringWithContext, Any] = field(default_factory=dict)
    aliases: List[str] = field(default_factory=list)
    is_grecord: bool = False
    is_pva: bool = False

    # For the convenience of downstream clients, redundantly keep track of the
    # associated IOC:
    owner: str = ""

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": textwrap.dedent(
            """\
            record({ record_type }}, "{{ name }}") {
            {% if owner %}
            {{ _indent }}# Part of {{ owner }}
            {% endif %}
            {% for ctx in context %}
            {{ _indent }}# {{ ctx }}
            {% endfor %}
            {% for name, field_inst in fields.items() | sort %}
            {{ _indent }}{{ render_object(field_inst, "console") | indent(_fmt.indent) }}
            {% endfor %}
            }
            """.rstrip()
        ),
        "file": textwrap.dedent(
            """\
            {% if is_grecord %}
            grecord({{ record_type }}, "{{ name }}") {
            {% else %}
            record({{ record_type }}, "{{ name }}") {
            {% endif %}
            {% for alias in aliases %}
            {{ _indent }}alias("{{ alias }}")
            {% endfor %}
            {% for name, field_inst in fields.items() | sort %}
            {{ _indent }}{{ render_object(field_inst, "file") | indent(_fmt.indent) }}
            {% endfor %}
            {% for name, info_value in info.items() | sort %}
            {{ _indent }}info({{ name }}, {{ _json_dump(info_value) | indent(_fmt.indent) }})
            {% endfor %}
            }
            """.rstrip()
        ),
    }

    @property
    def access_security_group(self) -> str:
        """The access security group name for the record."""
        if "ASG" in self.fields and not self.is_pva:
            return str(self.fields["ASG"].value)
        return "DEFAULT"

    def get_fields_of_type(self, *types) -> Generator[RecordField, None, None]:
        """Get all fields of the matching type(s)."""
        if self.is_pva:
            return

        for fld in self.fields.values():
            if fld.dtype in types:
                yield fld

    def get_links(
        self,
    ) -> Generator[Tuple[RecordField, str, List[str]], None, None]:
        """
        Get all links - in, out, and forward links - for this record.

        Requires that a dbd file was loaded when the record was created.
        Alternatively, see :func:`RecordType.get_links_for_record`.

        Yields
        ------
        field : RecordField
        link_text: str
        link_info: str
        """
        for fld in self.get_fields_of_type(*LINK_TYPES):
            try:
                link, info = get_link_information(fld.value)
            except ValueError:
                continue
            yield fld, link, info

    def to_summary(self) -> RecordInstanceSummary:
        """Return a summarized version of the record instance."""
        return RecordInstanceSummary.from_record_instance(self)

    def update(self, other: RecordInstance) -> List[LinterMessage]:
        """
        Update this record instance with another.

        TODO: This may not do exactly what an IOC would do.
        TODO: Return type?
        """
        if other.is_pva != self.is_pva:
            return [
                LinterError(
                    name="combine_pva_and_v3",
                    context=tuple(self.context) + tuple(other.context),
                    message="Cannot combine PVA group with V3 record"
                )
            ]
        self.context = remove_redundant_context(
            tuple(self.context) + tuple(other.context)
        )
        self.info.update(other.info)
        self.metadata.update(other.metadata)
        self.fields.update(other.fields)
        self.aliases.extend(
            [alias for alias in other.aliases if alias not in self.aliases]
        )

        self.has_dbd_info = self.has_dbd_info or other.has_dbd_info
        if self.record_type != other.record_type:
            return [
                LinterError(
                    name="record_type_mismatch",
                    context=self.context,
                    message=(
                        f"Record type mismatch in provided database files: "
                        f"{self.name} {self.record_type} {other.record_type}"
                    ),
                )
            ]

        return []


class AsynPortBase:
    """
    Base class for general asyn ports.

    Used in :mod:`whatrecord.asyn`, but made available here such that apischema
    can find it more readily.
    """
    _union: Any = None

    def __init_subclass__(cls, **kwargs):
        # Registers new subclasses automatically in the union cls._union.

        # Deserializers stack directly as a Union
        apischema.deserializer(
            apischema.conversions.Conversion(
                apischema.identity, source=cls, target=AsynPortBase
            )
        )

        # Only AsynPortBase serializer must be registered (and updated for each
        # subclass) as a Union, and not be inherited
        AsynPortBase._union = (
            cls if AsynPortBase._union is None else Union[AsynPortBase._union, cls]
        )
        apischema.serializer(
            apischema.conversions.Conversion(
                apischema.identity,
                source=AsynPortBase,
                target=AsynPortBase._union,
                inherited=False,
            )
        )


@dataclass
class ShellStateHandler:
    """A helper to work with interpreting commands from shell scripts."""
    metadata_key: ClassVar[str]
    parent: Optional[ShellStateHandler] = field(
        default=None, metadata=apischema.metadata.skip,
        repr=False, hash=False, compare=False
    )
    primary_handler: Optional[ShellState] = field(
        default=None, metadata=apischema.metadata.skip,
        repr=False, hash=False, compare=False
    )
    _handlers: Dict[str, Callable] = field(
        default_factory=dict, metadata=apischema.metadata.skip,
        repr=False, hash=False, compare=False, init=False
    )

    def __post_init__(self):
        self._handlers.update(dict(self.find_handlers()))
        self._init_sub_handlers_()

    def _init_sub_handlers_(self):
        """Initialize sub-handlers with parent/primary_handler."""
        for sub_handler in self.sub_handlers:
            sub_handler.parent = self
            sub_handler.primary_handler = getattr(self.parent, "primary_handler", self)
            sub_handler._init_sub_handlers_()

    def annotate_record(self, instance: RecordInstance) -> Optional[Dict[str, Any]]:
        """Annotate the given record's metadata with state-related information."""
        ...

    def get_load_context(self) -> FullLoadContext:
        """Get a FullLoadContext tuple representing where we are now."""
        if self.primary_handler is None:
            return tuple()
        return self.primary_handler.get_load_context()

    @property
    def sub_handlers(self) -> List[ShellStateHandler]:
        """Handlers which contain their own state."""
        return []

    @staticmethod
    def generic_handler_decorator(func=None, stub=False):
        """
        Decorate a handler method to generically return parameter-to-value
        information.

        This can be in addition to or in place of GDB command information.

        Parameters
        ----------
        func : callable
            The ``handler_`` method.

        stub : bool, optional
            Mark this as a stub method.  Variable arguments will be filled in
            as necessary, even if defaults are not provided.
        """

        def wrap(func):
            params = list(inspect.signature(func).parameters.items())[1:]
            defaults = list(
                None if param.default is inspect.Parameter.empty
                else param.default
                for _, param in params
            )

            @functools.wraps(func)
            def wrapped(self, *args):
                result = {}
                if len(args) < len(params) and stub:
                    # Pad unspecified arguments with defaults or "None"
                    args = list(args) + defaults[len(args):]

                if len(args) > len(params):
                    result["argument_lint"] = "Too many arguments"

                result["arguments"] = [
                    {
                        "name": name,
                        "type": getattr(param.annotation, "__name__", param.annotation),
                        "value": value,
                    }
                    for (name, param), value in zip(params, args)
                ]

                call_result = func(self, *args[:len(params)])
                if call_result is not None:
                    for key, value in call_result.items():
                        if key in result:
                            result[key].update(value)
                        else:
                            result[key] = value

                return result
            return wrapped

        if func is not None:
            return wrap(func)

        return wrap

    def find_handlers(self) -> Generator[Tuple[str, Callable], None, None]:
        """Find all IOC shell command handlers by name."""
        for handler_obj in [self] + self.sub_handlers:
            for attr in dir(handler_obj):
                if attr.startswith("handle_"):
                    obj = getattr(handler_obj, attr, None)
                    if callable(obj):
                        name = attr.split("_", 1)[1]
                        yield name, obj

            if handler_obj is not self:
                yield from handler_obj.find_handlers()

    def pre_ioc_init(self):
        """Pre-iocInit hook."""
        ...

    def post_ioc_init(self):
        """Post-iocInit hook."""
        ...


@dataclass
class RecordDefinitionAndInstance:
    """A pair of V3 record definition and instance."""
    definition: Optional[RecordType]
    instance: RecordInstance

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": """{{ render_object(instance, "console") }}""",
        "file": """{{ render_object(instance, "file") }}""",
    }


@dataclass
class WhatRecord:
    """
    WhatRecord - full set of information regarding a specific record.

    This response is on a per-IOC basis, so at most it can return one V3 record
    and one V4 record, as these exist in separate namespaces.


    Attributes
    ----------
    name : str
        The record name.

    record : RecordDefinitionAndInstance, optional
        The V3 record definition (if available) and record instance.

    pva_group : RecordInstance, optional
        The PVAccess group, if available.

    ioc : IocMetadata, optional
        The associated IOC metadata, if available.
    """
    name: str
    record: Optional[RecordDefinitionAndInstance] = None
    menus: Optional[Dict[str, DatabaseMenu]] = None
    pva_group: Optional[RecordInstance] = None
    ioc: Optional[IocMetadata] = None

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": textwrap.dedent(
            """\
            {{ name }}:
            {{ _indent }}{Owner: {{ present }}
            {{ _indent }}{IOC: {{ render_object(ioc, "console") }}
            {% if record %}
            {{ _indent }}{{ render_object(record, "console") | indent(_fmt.indent)}}
            {% endif %}
            {% if pva_group %}
            {{ _indent }}{{ render_object(pva_group, "console") | indent(_fmt.indent)}}
            {% endif %}
            }
            """.rstrip()
        ),
        "file": textwrap.dedent(
            """\
            {{ name }}:
            {{ _indent }}{Owner: {{ present }}
            {{ _indent }}{IOC: {{ render_object(ioc, "file") }}
            {% if record %}
            {{ _indent }}{{ render_object(record, "file") | indent(_fmt.indent)}}
            {% endif %}
            {% if pva_group %}
            {{ _indent }}{{ render_object(pva_group, "file") | indent(_fmt.indent)}}
            {% endif %}
            }
            """.rstrip(),
        ),
    }


@contextmanager
def time_context():
    """Return a callable to measure the time since context manager init."""
    start_count = perf_counter()

    def inner():
        return perf_counter() - start_count

    yield inner


def context_from_lark_token(fn: str, token: lark.Token) -> FullLoadContext:
    """Get a full load context from a given lark Token."""
    return (LoadContext(name=fn, line=token.line), )


def remove_redundant_context(full_context: FullLoadContext) -> FullLoadContext:
    """Remove redundant context information if it does not add anything."""
    if not full_context:
        return full_context

    # Inefficient, but the data set is small here, so meh
    zero_line_files = set(
        item.name
        for item in full_context
        if item.line == 0
    )

    for file in set(zero_line_files):
        for ctx in full_context:
            if ctx.name == file and ctx.line > 0:
                break
        else:
            zero_line_files.remove(file)

    new_context = []
    for ctx in full_context:
        is_specific = ctx.name not in zero_line_files or ctx.line > 0
        if is_specific and ctx not in new_context:
            new_context.append(ctx)
    return tuple(new_context)



import pathlib
import pprint

import apischema
import pytest

from ..autosave import AutosaveRestoreFile, RestoreError, RestoreValue
from ..common import LoadContext
from . import conftest

autosave_FILES = list((conftest.MODULE_PATH / "iocs").glob("**/*.sav"))

additional_files = conftest.MODULE_PATH / "autosave_filenames.txt"
if additional_files.exists():
    for additional in open(additional_files, "rt").read().splitlines():
        autosave_FILES.append(pathlib.Path(additional))


autosave_files = pytest.mark.parametrize(
    "autosave_file",
    [
        pytest.param(
            autosave_file,
            id="/".join(autosave_file.parts[-2:])
        )
        for autosave_file in autosave_FILES
    ]
)


@autosave_files
def test_parse(autosave_file):
    proto = AutosaveRestoreFile.from_file(autosave_file)
    serialized = apischema.serialize(proto)
    pprint.pprint(serialized)
    apischema.deserialize(AutosaveRestoreFile, serialized)


def test_basic():
    result = AutosaveRestoreFile.from_string(
        """\
# save/restore V5.1	Automatically generated - DO NOT MODIFY - 130618-005710
! 5 channel(s) not connected - or not all gets were successful
XPP:R30:EVR:27:CTRL.DG0C 119
<END>
""", filename="None")
    assert result.comments == [
        "save/restore V5.1	Automatically generated - DO NOT MODIFY - 130618-005710",
    ]
    assert result.values == {
        "XPP:R30:EVR:27:CTRL": {
            "DG0C": RestoreValue(
                context=(LoadContext("None", 3), ),
                record="XPP:R30:EVR:27:CTRL",
                field="DG0C",
                pvname="XPP:R30:EVR:27:CTRL.DG0C",
                value="119",
            )
        }
    }

    assert result.errors == [
        RestoreError(
            context=(LoadContext("None", 2), ),
            number=5,
            description="channel(s) not connected - or not all gets were successful",
        )
    ]


def test_basic_array():
    result = AutosaveRestoreFile.from_string(
        '''\
# save/restore V5.1	Automatically generated - DO NOT MODIFY - 130618-005710
! 5 channel(s) not connected - or not all gets were successful
XPP:R30:EVR:27:CTRL.DG0C @array@ { "500" "0.6" "0" "0" "0" "0" "0" "0" "0"'''
        + ''' "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" "0" }
<END>
''', filename="None"
    )
    assert result.comments == [
        "save/restore V5.1	Automatically generated - DO NOT MODIFY - 130618-005710",
    ]
    assert result.values == {
        "XPP:R30:EVR:27:CTRL": {
            "DG0C": RestoreValue(
                context=(LoadContext("None", 3), ),
                record="XPP:R30:EVR:27:CTRL",
                field="DG0C",
                pvname="XPP:R30:EVR:27:CTRL.DG0C",
                value=[
                    "500", "0.6", "0", "0", "0", "0", "0", "0", "0", "0", "0",
                    "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0",
                    "0", "0"
                ]
            )
        }
    }

    assert result.errors == [
        RestoreError(
            context=(LoadContext("None", 2), ),
            number=5,
            description="channel(s) not connected - or not all gets were successful",
        )
    ]


def test_basic_empty_value():
    result = AutosaveRestoreFile.from_string(
        """XPP:R30:EVR:27:CTRL.DG0C \n<END>""", filename="None"
    )
    assert result.values == {
        "XPP:R30:EVR:27:CTRL": {
            "DG0C": RestoreValue(
                context=(LoadContext("None", 1), ),
                record="XPP:R30:EVR:27:CTRL",
                field="DG0C",
                pvname="XPP:R30:EVR:27:CTRL.DG0C",
                value="",
            )
        }
    }


def test_basic_string():
    result = AutosaveRestoreFile.from_string(
        r"""
XPP:R30:EVR:27:CTRL.DG0C "a quoted string"
<END>
""", filename="None"
    )
    assert result.values == {
        "XPP:R30:EVR:27:CTRL": {
            "DG0C": RestoreValue(
                context=(LoadContext("None", 2),),
                record="XPP:R30:EVR:27:CTRL",
                field="DG0C",
                pvname="XPP:R30:EVR:27:CTRL.DG0C",
                value="a quoted string",
            )
        }
    }


def test_basic_escaped():
    result = AutosaveRestoreFile.from_string(
        r"""
CTRL.DG0C a \"quoted\" string
CTRL.DG1C @array@ { "1.23" " 2.34" " 3.45" }
CTRL.DG2C @array@ { "abc" "de\"f" "g{hi\"" "jkl mno} pqr" }
<END>
""", filename="None")
    assert result.values == {
        "CTRL": {
            "DG0C": RestoreValue(
                context=(LoadContext("None", 2), ),
                record="CTRL",
                field="DG0C",
                pvname="CTRL.DG0C",
                value='a "quoted" string',
            ),
            "DG1C": RestoreValue(
                context=(LoadContext("None", 3), ),
                record="CTRL",
                field="DG1C",
                pvname="CTRL.DG1C",
                value=["1.23", " 2.34", " 3.45"],
            ),
            "DG2C": RestoreValue(
                context=(LoadContext("None", 4), ),
                record="CTRL",
                field="DG2C",
                pvname="CTRL.DG2C",
                value=["abc", 'de"f', 'g{hi"', "jkl mno} pqr"],
            ),
        }
    }


def test_basic_disconnected():
    result = AutosaveRestoreFile.from_string(
        r"""
XPP:R30:EVR:27:CTRL.DG0C a \"quoted\" string
#abc Search Issued
#def Search Issued
XPP:R30:EVR:27:CTRL.DG1C an unquoted string
#ghi Search Issued
<END>
""")
    assert result.disconnected == [
        "abc",
        "def",
        "ghi",
    ]



"""Smoke tests to see if the provided IOCs load without crashing."""

import os
import pathlib

import apischema
import pytest

from ..bin.parse import main
from ..parse import parse
from ..shell import LoadedIoc, ShellState
from . import conftest

empty_db_iocs = {"fake_ad", "ioc_failure"}


@conftest.startup_scripts
def test_load_smoke(startup_script):
    os.environ["PWD"] = str(startup_script.resolve().parent)
    loaded_ioc: LoadedIoc = parse(startup_script)
    if startup_script.parent.name not in empty_db_iocs:
        assert loaded_ioc.shell_state.database
        assert loaded_ioc.shell_state.database_definition


@conftest.startup_scripts
def test_load_round_trip_smoke(startup_script):
    os.environ["PWD"] = str(startup_script.resolve().parent)
    loaded_ioc = parse(startup_script)
    apischema.deserialize(LoadedIoc, apischema.serialize(loaded_ioc, ))

    assert loaded_ioc.metadata.is_up_to_date()
    # assert loaded_ioc == round_tripped


@conftest.startup_scripts
def test_load_and_whatrec(startup_script):
    os.environ["PWD"] = str(startup_script.resolve().parent)
    loaded_ioc: LoadedIoc = parse(startup_script)
    state: ShellState = loaded_ioc.shell_state

    v3_db = state.database
    if v3_db:
        pvname = list(v3_db)[0]
        whatrec = loaded_ioc.whatrec(pvname)
        assert whatrec is not None

        rec = whatrec.record.instance
        dbd = whatrec.record.definition
        assert whatrec.name == rec.name == pvname
        assert rec == state.database[pvname]
        if rec.record_type != "motor":  # sorry, dbd doesn't have it
            assert dbd == state.database_definition.record_types[rec.record_type]

    pva_db = state.pva_database
    if pva_db:
        pvname = list(pva_db)[0]
        whatrec = loaded_ioc.whatrec(pvname)
        assert whatrec is not None

        rec = whatrec.pva_group
        assert whatrec.name == rec.name == pvname
        assert rec == state.pva_database[pvname]


output_formats = pytest.mark.parametrize(
    "output_format",
    [
        "json",
        "console",
        "file",
    ],
)


@conftest.startup_scripts
@output_formats
def test_load_dump(startup_script: pathlib.Path, output_format: str):
    os.environ["PWD"] = str(startup_script.resolve().parent)
    main(startup_script, output_format=output_format)


@output_formats
@pytest.mark.parametrize(
    "filename",
    [
        pytest.param(conftest.MODULE_PATH / fn, id=fn)
        for fn in [
            "access.acf",
            "kfe.pvlist",
            "iocs/db/test.substitutions",
            "iocs/streamdevice/hex.proto",
        ]
    ]
)
def test_load_misc(filename: pathlib.Path, output_format: str):
    os.environ["PWD"] = str(filename.resolve().parent)
    main(filename, output_format=output_format)



"""
LDAP Configuration summary plugin
"""

from __future__ import annotations

import argparse
import json
import logging
import os
import typing
from dataclasses import dataclass
from typing import Dict, List

import apischema
import ldap

from ..server.common import PluginResults

# from .util import suppress_output_decorator

logger = logging.getLogger(__name__)

# Stash the description for later usage by the CLI interface.
DESCRIPTION = __doc__.strip()

# LDAP base
# dc = Domain Component (reg)
# o = Organization (slac)
# ou = Organization unit (Subnets)
LDAP_BASE = os.environ.get("WHATRECORD_LDAP_BASE", "ou=Subnets,dc=reg,o=slac")

# LDAP scope (subtree, base, onelevel) - this probably remains as-is
LDAP_SCOPE = os.environ.get("WHATRECORD_LDAP_SCOPE", "subtree")

# LDAP filter
LDAP_FILTER = os.environ.get("WHATRECORD_LDAP_FILTER", "cn=*")

# The LDAP server URI
LDAP_SERVER = os.environ.get("WHATRECORD_LDAP_SERVER", "ldap://psldap1.pcdsn/")

# The LDAP byte encoding
LDAP_ENCODING = os.environ.get("WHATRECORD_LDAP_ENCODING", "utf-8")


@dataclass
class NetconfigPluginResults(PluginResults):
    # Could potentially further specify metadata_by_key or metadata
    ...


def decode_ldap_attrs(
    attrs: Dict[str, List[bytes]], encoding=LDAP_ENCODING
) -> Dict[str, List[str]]:
    """Decode byte-filled LDAP attrs dictionary."""
    def decode_value(key: str, value: List[bytes]):
        return [part.decode(encoding, "replace") for part in value]

    return {key: decode_value(key, value) for key, value in attrs.items()}


def main(
    server: str = LDAP_SERVER,
    base: str = LDAP_BASE,
    scope: int = ldap.SCOPE_SUBTREE,
    filter_query: str = LDAP_FILTER,
    encoding: str = LDAP_ENCODING,
    pretty: bool = False,
):
    client = ldap.initialize(server)
    metadata = {}
    for dn, attrs in client.search_s(base, scope, filter_query):
        info = decode_ldap_attrs(attrs, encoding=encoding)
        for cn in info["cn"]:
            cn_info = metadata.setdefault(cn, {})
            for key, value in info.items():
                cn_info.setdefault(key, []).extend(value)
            for dninfo in ldap.dn.str2dn(dn):
                for key, value, *_ in dninfo:
                    cn_info.setdefault(key, []).append(value)

    return NetconfigPluginResults(
        files_to_monitor={},
        record_to_metadata_keys={},
        metadata=metadata,
        metadata_by_key={},
        execution_info={},
    )


def _get_argparser(
    parser: typing.Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    if parser is None:
        parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        "--server",
        type=str,
        help="LDAP server URI",
        default=LDAP_SERVER,
    )

    parser.add_argument(
        "--base",
        type=str,
        help="LDAP base",
        default=LDAP_BASE,
    )

    parser.add_argument(
        "--scope",
        type=str,
        help="LDAP scope",
        choices=("subtree", "base", "onelevel"),
        default=LDAP_SCOPE,
    )

    parser.add_argument(
        "--filter",
        type=str,
        dest="filter_query",
        help="LDAP filter / search query",
        default=LDAP_FILTER,
    )

    parser.add_argument(
        "--encoding",
        type=str,
        help="LDAP string encoding",
        default=LDAP_ENCODING,
    )

    parser.add_argument(
        "-p", "--pretty", action="store_true", help="Pretty JSON output"
    )
    return parser


def _cli_main():
    parser = _get_argparser()
    kwargs = vars(parser.parse_args())
    kwargs["scope"] = {
        "subtree": ldap.SCOPE_SUBTREE,
        "base": ldap.SCOPE_BASE,
        "onelevel": ldap.SCOPE_ONELEVEL,
    }[kwargs["scope"]]
    results = main(**kwargs)
    json_results = apischema.serialize(results)
    dump_args = {"indent": 4} if kwargs["pretty"] else {}
    print(json.dumps(json_results, sort_keys=True, **dump_args))


if __name__ == "__main__":
    _cli_main()



from __future__ import annotations

import collections
import logging
import pathlib
import shlex
import textwrap
from dataclasses import dataclass, field
from typing import List, Optional, Sequence, Union

import graphviz as gv
import lark

from . import transformer
from .common import AnyPath, FullLoadContext
from .graph import _GraphHelper
from .transformer import context_from_token

logger = logging.getLogger(__name__)


@dataclass
class Definition:
    context: FullLoadContext


@dataclass
class Expression:
    context: FullLoadContext


@dataclass
class CommaSeparatedExpressions(Expression):
    expressions: List[Expression]

    def __str__(self) -> str:
        return ", ".join(str(item) for item in self.expressions)


OptionalExpression = Optional[Union[Expression, CommaSeparatedExpressions]]


@dataclass
class Assignment(Definition):
    variable: str
    value: Optional[Union[str, Sequence[str]]] = None
    subscript: Optional[str] = None

    def __str__(self) -> str:
        if isinstance(self.value, tuple):
            value = "{ " + ", ".join(str(s) for s in self.value) + "}"
        else:
            value = str(self.value)
        return f"{self.variable}{self.subscript or ''} = {value};"


@dataclass
class AbstractDeclaratorModifier:
    context: FullLoadContext
    modifier: str
    decl: Optional[AbstractDeclarator] = None

    def __str__(self) -> str:
        if self.modifier == "(":  # )
            return f"({self.decl})"
        if self.modifier == "*":
            return f"*{self.decl or ''}"
        return f"{self.modifier} {self.decl or ''}"


@dataclass
class AbstractDeclarator:
    context: FullLoadContext
    params: Optional[Sequence[ParameterDeclarator]] = None
    subscript: Optional[str] = None
    inner_decl: Optional[AbstractDeclarator] = None

    def __str__(self) -> str:
        prefix = f"{str(self.inner_decl).strip()} " if self.inner_decl is not None else ""
        if self.subscript is not None:
            return f"{prefix}{self.subscript}"
        if self.params is not None:
            params = ", ".join(str(param) for param in self.params)
            return f"{prefix}({params})"
        return prefix.strip()


@dataclass
class Type:
    context: FullLoadContext
    name: str
    abstract: Optional[AbstractDeclarator] = None

    def __str__(self) -> str:
        return f"{self.abstract or ''}{self.name}"


@dataclass
class Monitor(Definition):
    variable: str
    subscript: Optional[int]

    def __str__(self) -> str:
        return f"monitor {self.variable}{self.subscript or ''};"


@dataclass
class Option(Definition):
    name: str
    enable: bool

    def __str__(self) -> str:
        plus = "+" if self.enable else "-"
        return f"option {plus}{self.name};"


@dataclass
class Sync(Definition):
    variable: str
    subscript: Optional[int]
    queued: bool
    event_flag: Optional[str] = None
    queue_size: Optional[int] = None

    def __str__(self) -> str:
        subscript = str(self.subscript or "")
        variable = f"{self.variable}{subscript}"
        event_flag = f" to {self.event_flag}" if self.event_flag else ""
        if self.queued:
            return f"syncq {variable}{event_flag} {self.queue_size or ''};"

        return f"sync {variable}{event_flag};"


@dataclass
class Declaration(Definition):
    type: Optional[Type] = None
    declarators: Optional[Sequence[Declarator]] = field(default_factory=list)

    def __str__(self) -> str:
        declarators = ", ".join(str(decl) for decl in self.declarators or [])
        return f"{self.type} {declarators};"


@dataclass
class ForeignDeclaration(Declaration):
    names: Sequence[str] = field(default_factory=list)

    def __str__(self) -> str:
        names = ", ".join(self.names)
        return f"foreign {names};"


@dataclass
class Declarator:
    context: FullLoadContext
    object: Union[Declarator, Variable]
    params: Optional[List[ParameterDeclarator]] = None
    value: Optional[Expression] = None
    modifier: Optional[str] = None
    subscript: Optional[int] = None

    def __str__(self) -> str:
        if self.modifier == "()":
            decl = f"({self.object})"
        elif self.modifier:
            decl = f"{self.modifier}{self.object}"
        elif self.subscript:
            decl = f"{self.object}{self.subscript}"
        elif self.params:
            params = ", ".join(str(param) for param in self.params)
            decl = f"{self.object}({params})"
        else:
            decl = str(self.object)
        if self.value:
            return f"{decl} = {self.value}"
        return f"{decl}"


@dataclass
class ParameterDeclarator:
    context: FullLoadContext
    type: Type
    declarator: Optional[Declarator] = None

    def __str__(self) -> str:
        if self.declarator:
            return f"{self.type} {self.declarator}"
        return str(self.type)


@dataclass
class State:
    context: FullLoadContext
    name: str
    definitions: Sequence[Definition] = field(default_factory=list)
    transitions: Sequence[Transition] = field(default_factory=list)
    entry: Optional[Block] = None
    exit: Optional[Block] = None
    code_transitions: List[StateStatement] = field(default_factory=list)

    def __str__(self) -> str:
        definitions = "\n".join(
            str(defn)
            for defn in self.definitions
        )
        transitions = "\n".join(
            str(transition)
            for transition in self.transitions
        )
        return "\n".join(
            line for line in (
                f"state {self.name}" + " {",
                str(self.entry or ""),
                textwrap.indent(definitions, " " * 4),
                textwrap.indent(transitions, " " * 4),
                str(self.exit or ""),
                "}",
            )
            if line
        )


@dataclass
class StateSet:
    context: FullLoadContext
    name: str
    definitions: Sequence[Definition] = field(default_factory=list)
    states: Sequence[State] = field(default_factory=list)

    def __str__(self) -> str:
        definitions = "\n".join(
            str(defn)
            for defn in self.definitions
        )
        states = "\n".join(
            str(state)
            for state in self.states
        )
        return "\n".join(
            (
                f"ss {self.name}" + " {",
                textwrap.indent(definitions, " " * 4),
                textwrap.indent(states, " " * 4),
                "}",
            )
        )


@dataclass
class Transition:
    context: FullLoadContext
    block: Block
    target_state: Optional[str] = None
    condition: OptionalExpression = None

    def __str__(self) -> str:
        state = (
            f"state {self.target_state}"
            if self.target_state
            else
            "exit"
        )
        return "\n".join(
            line for line in (
                f"when ({self.condition or ''})" + " {",
                str(self.block),
                "}" + f" {state}",
            )
            if line
        )


@dataclass
class ExitTransition(Transition):
    ...


@dataclass
class Block:
    context: FullLoadContext
    definitions: Sequence[Definition] = field(default_factory=list)
    statements: Sequence[Statement] = field(default_factory=list)

    def __str__(self) -> str:
        definitions = textwrap.indent(
            "\n".join(
                str(defn)
                for defn in self.definitions
            ),
            prefix="    ",
        )
        statements = textwrap.indent(
            "\n".join(
                str(statement)
                for statement in self.statements
            ),
            prefix="    ",
        )
        return "\n".join(
            (
                "{",
                definitions,
                statements,
                "}",
            )
        )


@dataclass
class Statement:
    context: FullLoadContext


@dataclass
class BreakStatement(Statement):
    def __str__(self) -> str:
        return "break;"


@dataclass
class ContinueStatement(Statement):
    def __str__(self) -> str:
        return "continue;"


@dataclass
class ReturnStatement(Statement):
    value: Optional[Expression] = None

    def __str__(self) -> str:
        if self.value is not None:
            return f"return {self.value};"
        return "return;"


@dataclass
class StateStatement(Statement):
    name: str

    def __str__(self) -> str:
        return f"state {self.name};"


@dataclass
class WhileStatement(Statement):
    condition: CommaSeparatedExpressions
    body: Statement

    def __str__(self) -> str:
        return f"while ({self.condition}) " + str(self.body).lstrip()


@dataclass
class ForStatement(Statement):
    init: OptionalExpression
    condition: OptionalExpression
    increment: OptionalExpression
    statement: Statement

    def __str__(self) -> str:
        return (
            f"for ({self.init or ''}; {self.condition or ''}; {self.increment or ''}) " +
            str(self.statement).lstrip()
        )


@dataclass
class ExpressionStatement(Statement):
    expression: CommaSeparatedExpressions

    def __str__(self) -> str:
        return f"{self.expression};"


@dataclass
class IfStatement(Statement):
    condition: CommaSeparatedExpressions
    body: Statement
    else_body: Optional[Statement] = None

    def __str__(self) -> str:
        else_clause = "\n".join(
            (
                "\nelse {",
                textwrap.indent(str(self.else_body), " " * 4),
                "}",
            )
            if self.else_body else ""
        )

        return "\n".join(
            (
                f"if ({self.condition}) " + " {",
                textwrap.indent(str(self.body), " " * 4),
                else_clause,
                "}",
            )
        )


@dataclass
class FuncDef(Definition):
    type: Type
    declarator: Declarator
    block: Block

    def __str__(self) -> str:
        return f"{self.type} {self.declarator};"


@dataclass
class StructMemberDecl:
    context: FullLoadContext
    type: Type
    declarator: Declarator

    def __str__(self) -> str:
        return f"{self.type} {self.declarator};"


@dataclass
class CCode(Definition):
    code: str

    def __str__(self) -> str:
        return self.code


@dataclass
class StructDef(Definition):
    name: str
    members: Sequence[Union[StructMemberDecl, CCode]] = field(default_factory=list)

    def __str__(self) -> str:
        members = "\n".join(
            str(member) for member in self.members
        )
        return "\n".join(
            (
                f"struct {self.name}" + " {",
                textwrap.indent(members, " " * 4),
                "}",
            )
        )


@dataclass
class Variable(Expression):
    name: str

    def __str__(self) -> str:
        return self.name


@dataclass
class InitExpression(Expression):
    """
    Of the form:
        ( type ) { init_exprs }
        { init_exprs }
        expr
    """

    # TODO: may be improved?
    context: FullLoadContext
    expressions: Sequence[Union[InitExpression, Expression]] = field(
        default_factory=list
    )
    type: Optional[Type] = None

    def __str__(self) -> str:
        exprs = ", ".join(str(expr or '') for expr in self.expressions)
        type_prefix = f"({self.type}) " if self.type else ""
        return f"{type_prefix}{{ {exprs} }}"


@dataclass
class Literal(Expression):
    type: str
    value: str

    def __str__(self) -> str:
        return self.value


@dataclass
class UnaryPrefixExpression(Expression):
    operator: str
    expression: Expression

    def __str__(self) -> str:
        return f"{self.operator}{self.expression}"


@dataclass
class UnaryPostfixExpression(Expression):
    expression: Expression
    operator: str

    def __str__(self) -> str:
        return f"{self.expression}{self.operator}"


@dataclass
class BinaryOperatorExpression(Expression):
    left: Expression
    operator: str
    right: Expression

    def __str__(self) -> str:
        return f"{self.left} {self.operator} {self.right}"


@dataclass
class TypeCastExpression(Expression):
    type: Type
    expression: Expression

    def __str__(self) -> str:
        return f"{self.type}({self.expression})"


@dataclass
class TernaryExpression(Expression):
    condition: Expression
    if_true: Expression
    if_false: Expression

    def __str__(self) -> str:
        return f"{self.condition} ? {self.if_true} : {self.if_false}"


@dataclass
class MemberExpression(Expression):
    parent: Expression
    member: str
    dereference: bool  # True = ->, False = .

    def __str__(self) -> str:
        if self.dereference:
            return f"{self.parent}->{self.member}"
        return f"{self.parent}.{self.member}"


@dataclass
class SizeofExpression(Expression):
    type: Type

    def __str__(self) -> str:
        return f"sizeof({self.type})"


@dataclass
class ExitExpression(Expression):
    args: OptionalExpression

    def __str__(self) -> str:
        return f"exit({self.args or ''})"


@dataclass
class BracketedExpression(Expression):
    outer: Expression
    inner: Expression

    def __str__(self) -> str:
        return f"{self.outer}[{self.inner}]"


@dataclass
class ParenthesisExpression(Expression):
    expression: Expression

    def __str__(self) -> str:
        return f"({self.expression})"


@dataclass
class ExpressionWithArguments(Expression):
    expression: Expression
    arguments: OptionalExpression = None

    def __str__(self) -> str:
        return f"{self.expression}({self.arguments or ''})"


class SequencerProgramGraph(_GraphHelper):
    """
    A graph for a SequencerProgram.

    Parameters
    ----------
    program : SequencerProgram, optional
        A program to add.

    highlight_states : list of str, optional
        List of state names to highlight.

    include_code : bool, optional
        Include code, where relevant, in nodes.
    """

    _entry_label: str = "_Entry_"
    _exit_label: str = "_Exit_"

    def __init__(
        self,
        program: Optional[SequencerProgram] = None,
        highlight_states: Optional[List[str]] = None,
        include_code: bool = False,
    ):
        super().__init__()
        self.include_code = include_code
        self.highlight_states = highlight_states or []
        if program is not None:
            self.add_program(program)

    def add_program(self, program: SequencerProgram):
        """Add a program to the graph."""
        for state_set in program.state_sets:
            for state in state_set.states:
                self._add_state(program, state_set, state)

        # Only add entry/exit labels if there's an edge
        if self._entry_label in self.nodes:
            self.get_node(
                self._entry_label,
                self.get_code(program.entry, "(Startup)")
            )

        if self._exit_label in self.nodes:
            self.get_node(
                self._exit_label,
                self.get_code(program.exit, "(Exit)")
            )

    def _add_state(self, program: SequencerProgram, state_set: StateSet, state: State):
        """Add a state set's state to the graph."""
        qualified_name = f"{state_set.name}.{state.name}"
        self.get_node(
            qualified_name,
            text="\n".join(self._get_state_node_text(state_set, state)),
        )

        if state_set.states[0] is state:
            self.add_edge(
                self._entry_label,
                qualified_name,
            )

        for transition in state.transitions:
            self._add_transition(program, state_set, state, transition)

        for transition in state.code_transitions:
            self.add_edge(
                qualified_name,
                f"{state_set.name}.{transition.name}",
                label=f"(Line {transition.context[-1].line})",
            )

    def _ready_for_digraph(self, graph: gv.Digraph):
        """Hook when the user calls ``to_digraph``."""
        for node in self.nodes.values():
            node.highlighted = node.label in self.highlight_states

    def _add_transition(
        self,
        program: SequencerProgram,
        state_set: StateSet,
        state: State,
        transition: Transition,
    ):
        """Add a state set's state transition to the graph."""
        state_qualified_name = f"{state_set.name}.{state.name}"
        label = str(transition.condition or "")
        transition_text = self.get_code(transition.block)
        target_state = (
            f"{state_set.name}.{transition.target_state}"
            if transition.target_state
            else self._exit_label
        )
        if not self.include_code or not transition_text:
            self.add_edge(state_qualified_name, target_state, label=label)
            return

        transition_idx = state.transitions.index(transition)
        transition_node = f"{state_qualified_name}.{transition_idx}"
        self.get_node(transition_node, text=transition_text)
        self.add_edge(state_qualified_name, transition_node, label=label)
        self.add_edge(transition_node, target_state)

    def get_code(self, obj, default: str = ""):
        """Get code for a node/edge."""
        if self.include_code and obj is not None:
            return self.clean_code(str(obj)) or default
        return default

    def _get_state_node_text(self, state_set: StateSet, state: State):
        yield f"<b>{state_set.name}.{state.name}</b>"
        if self.include_code:
            if state.entry is not None:
                yield "<u>Entry</u>"
                yield self.clean_code(state.entry)
            if state.exit is not None:
                yield "<u>Exit</u>"
                yield self.clean_code(state.exit)


@dataclass
class SequencerProgram:
    """Representation of a state notation language (snl seq) program."""

    context: FullLoadContext
    name: str
    params: Optional[str]
    initial_definitions: Sequence[Definition] = field(default_factory=list)
    entry: Optional[Block] = None
    state_sets: Sequence[StateSet] = field(default_factory=list)
    exit: Optional[Block] = None
    final_definitions: Sequence[Definition] = field(default_factory=list)

    def __str__(self) -> str:
        initial_definitions = "\n".join(
            str(defn)
            for defn in self.initial_definitions
        )
        final_definitions = "\n".join(
            str(defn)
            for defn in self.final_definitions
        )
        state_sets = "\n\n".join(
            str(state_set)
            for state_set in self.state_sets
        )
        param = f"({self.params})" if self.params else ""
        return "\n\n".join(
            line for line in (
                f"program {self.name}{param}",
                initial_definitions,
                str(self.entry or ""),
                state_sets,
                str(self.exit or ""),
                final_definitions,
            )
            if line
        )

    @staticmethod
    def preprocess(code: str, search_path: Optional[AnyPath] = None) -> str:
        """Preprocess the given sequencer code, expanding #include."""
        # Line numbers will be off with this, sadly
        # The sequencer itself gets around this by using LINE_MARKER tokens
        # to indicate what file and line the code came from.  This could
        # be something we support in the future, but it might not be easy
        # with lark...
        search_path = pathlib.Path("." if search_path is None else search_path)
        result = []
        stack = collections.deque([(search_path, line) for line in code.splitlines()])
        while stack:
            search_path, line = stack.popleft()
            if line.startswith("#include"):
                _, include_file, *_ = shlex.split(line)
                include_file = (search_path / include_file).resolve()
                with open(include_file, "rt") as fp:
                    stack.extendleft(
                        [
                            (include_file.parent, line)
                            for line in reversed(fp.read().splitlines())
                        ]
                    )
            elif line.startswith("#if"):
                ...  # sorry; this may break things
            elif line.startswith("#else"):
                ...  # sorry; this may break things
            elif line.startswith("#elif"):
                ...  # sorry; this may break things
            elif line.startswith("#endif"):
                ...  # sorry; this may break things
            elif line.startswith("#define"):
                while stack and line.endswith("\\"):
                    search_path, line = stack.popleft()

                ...  # sorry; I think we can do better
            else:
                result.append(line)

        return "\n".join(result)

    @classmethod
    def from_string(
        cls,
        contents: str,
        filename: Optional[AnyPath] = None,
        debug: bool = False,
    ) -> SequencerProgram:
        """Load a state notation language file given its string contents."""
        comments = []
        grammar = lark.Lark.open_from_package(
            "whatrecord",
            "snl.lark",
            search_paths=("grammar",),
            parser="earley",
            # TODO: alternative comment finding method
            # lexer_callbacks={"COMMENT": comments.append},
            maybe_placeholders=True,
            propagate_positions=True,
            debug=debug,
        )

        search_path = None
        if filename:
            search_path = pathlib.Path(filename).resolve().parent

        preprocessed = cls.preprocess(contents, search_path=search_path)
        proto = _ProgramTransformer(cls, filename).transform(
            grammar.parse(preprocessed)
        )
        proto.comments = comments
        return proto

    @classmethod
    def from_file_obj(cls, fp, filename: Optional[AnyPath] = None) -> SequencerProgram:
        """Load a state notation language program given a file object."""
        return cls.from_string(
            fp.read(),
            filename=getattr(fp, "name", filename),
        )

    @classmethod
    def from_file(cls, fn: AnyPath) -> SequencerProgram:
        """
        Load a state notation language file.

        Parameters
        ----------
        filename : pathlib.Path or str
            The filename.

        Returns
        -------
        program : SequencerProgram
            The parsed program.
        """
        with open(fn, "rt") as fp:
            return cls.from_string(fp.read(), filename=fn)

    def as_graph(self, **kwargs) -> SequencerProgramGraph:
        """
        Create a graphviz digraph of the state notation diagram.

        Returns
        -------
        graph : SequencerProgramGraph
        """
        return SequencerProgramGraph(self, **kwargs)


@lark.visitors.v_args(inline=True)
class _ProgramTransformer(lark.visitors.Transformer):
    def __init__(self, cls, fn, visit_tokens=False):
        super().__init__(
            visit_tokens=visit_tokens,
        )
        self.fn = str(fn)
        self.cls = cls
        self._code_transitions = []

    def program(
        self,
        program_token: lark.Token,
        name: lark.Token,
        program_param: Optional[str],
        initial_defns: Sequence[Definition],
        entry: Block,
        state_sets: Sequence[StateSet],
        exit: Block,
        final_defns: Sequence[Definition],
    ):
        """
        PROGRAM NAME program_param initial_defns entry state_sets exit final_defns
        """
        return self.cls(
            context=context_from_token(self.fn, program_token),
            params=program_param,
            name=str(name),
            initial_definitions=initial_defns,
            entry=entry,
            state_sets=state_sets,
            exit=exit,
            final_definitions=final_defns,
        )

    def program_param(self, *args):
        if not args:
            return None
        """
        LPAREN string RPAREN
        """
        return str(args[1])

    initial_defns = transformer.tuple_args
    final_defns = transformer.tuple_args

    def assign(self, assign_token, variable, _):
        """
        ASSIGN variable SEMICOLON
        """
        return Assignment(
            context=context_from_token(self.fn, assign_token),
            variable=str(variable),
        )

    def assign_string(self, assign_token, variable, to_, value, _):
        """
        ASSIGN variable to string SEMICOLON
        """
        return Assignment(
            context=context_from_token(self.fn, assign_token),
            variable=str(variable),
            value=str(value),
        )

    def assign_subscript_string(self, assign_token, variable, subscript, to_, value, _):
        """
        ASSIGN variable subscript to string SEMICOLON
        """
        return Assignment(
            context=context_from_token(self.fn, assign_token),
            variable=str(variable),
            value=str(value),
            subscript=str(subscript),
        )

    def assign_strings(self, assign_token, variable, to_, _, strings, *__):
        """
        ASSIGN variable to LBRACE strings RBRACE SEMICOLON
        """
        return Assignment(
            context=context_from_token(self.fn, assign_token),
            variable=str(variable),
            value=tuple(str(value) for value in strings),
        )

    strings = transformer.tuple_args

    def monitor(self, monitor_token, variable, opt_subscript, _):
        """
        MONITOR variable opt_subscript SEMICOLON
        """
        return Monitor(
            context=context_from_token(self.fn, monitor_token),
            variable=str(variable),
            subscript=str(opt_subscript) if opt_subscript else None,
        )

    def sync(self, sync_token, variable, subscript, _, event_flag, __):
        """
        SYNC variable opt_subscript to event_flag SEMICOLON
        """
        return Sync(
            context=context_from_token(self.fn, sync_token),
            variable=str(variable),
            subscript=str(subscript) if subscript else None,
            event_flag=event_flag,
            queued=False,
        )

    def syncq_flagged(
        self, syncq_token, variable, subscript, _, event_flag, syncq_size, __
    ):
        """
        SYNCQ variable opt_subscript to event_flag syncq_size SEMICOLON
        """
        return Sync(
            context=context_from_token(self.fn, syncq_token),
            variable=str(variable),
            subscript=str(subscript) if subscript else None,
            event_flag=event_flag,
            queued=True,
            queue_size=syncq_size,
        )

    def syncq(self, syncq_token, variable, subscript, syncq_size, _):
        """
        SYNCQ variable opt_subscript syncq_size SEMICOLON
        """
        return Sync(
            context=context_from_token(self.fn, syncq_token),
            variable=str(variable),
            subscript=str(subscript) if subscript else None,
            queued=True,
            queue_size=syncq_size,
        )

    event_flag = transformer.stringify
    variable = transformer.pass_through

    def syncq_size(self, size=None):
        """
        INTCON?
        """
        return int(size) if size else None

    opt_subscript = transformer.pass_through

    def subscript(self, _, value, __):
        """
        LBRACKET INTCON RBRACKET
        """
        return value

    def declaration(self, basetype: Type, init_declarators, _):
        """
        basetype init_declarators SEMICOLON
        """
        return Declaration(
            context=basetype.context,
            type=basetype,
            declarators=init_declarators,
        )

    def foreign_declaration(self, foreign_token, variables, _):
        """
        FOREIGN variables SEMICOLON
        """
        return ForeignDeclaration(
            context=context_from_token(self.fn, foreign_token),
            names=tuple(str(variable) for variable in variables),
            type=None,
            declarators=None,
        )

    init_declarators = transformer.tuple_args

    def init_declarator(self, declarator: Declarator, *equal_expr):
        if equal_expr:
            _, value = equal_expr
            declarator.value = value
        return declarator

    def declarator(self, variable: lark.Token):
        return Declarator(
            context=context_from_token(self.fn, variable),
            object=Variable(
                context=context_from_token(self.fn, variable),
                name=str(variable),
            ),
        )

    def declarator_decls(self, declarator, _, param_decls, __):
        return Declarator(
            context=declarator.context,
            object=declarator,
            params=param_decls,
        )

    def declarator_subscript(self, declarator, subscript):
        return Declarator(
            context=declarator.context,
            object=declarator,
            subscript=str(subscript),
        )

    def declarator_paren(self, _, declarator, __):
        return Declarator(
            context=declarator.context,
            object=declarator,
            modifier="()",
        )

    def declarator_deref(self, asterisk, declarator):
        return Declarator(
            context=context_from_token(self.fn, asterisk),
            object=declarator,
            modifier="*",
        )

    def declarator_const(self, const, declarator):
        return Declarator(
            context=context_from_token(self.fn, const),
            object=declarator,
            modifier="const",
        )

    param_decls = transformer.tuple_args

    def param_decl(self, type_, declarator=None):
        """
        basetype declarator
        type_expr
        """
        return ParameterDeclarator(
            context=type_.context, type=type_, declarator=declarator
        )

    variables = transformer.tuple_args

    init_expr = transformer.pass_through

    def typed_init_expr(self, lparen, type_expr, _, __, init_exprs, ___):
        """
        type_expr RPAREN LBRACE init_exprs RBRACE
        LBRACE init_exprs RBRACE
        """
        return InitExpression(
            context=context_from_token(self.fn, lparen),
            expressions=init_exprs,
            type=type_expr,
        )

    def untyped_init_expr(self, lbrace, init_exprs, _):
        """
        LBRACE init_exprs RBRACE
        """
        return InitExpression(
            context=context_from_token(self.fn, lbrace),
            expressions=init_exprs,
            type=None,
        )

    init_exprs = transformer.tuple_args

    def basetype(self, *type_info):
        return Type(
            context=context_from_token(self.fn, type_info[0]),
            name=" ".join(type_info),
        )

    def type_expr(self, basetype: Type, abs_decl: Optional[AbstractDeclarator] = None):
        basetype.abstract = abs_decl
        return basetype

    def abs_decl_mod(
        self, token: lark.Token, abs_decl: Optional[AbstractDeclarator] = None, *_
    ):
        """
        LPAREN abs_decl RPAREN
        ASTERISK [ abs_decl ]
        CONST [ abs_decl ]
        """
        return AbstractDeclaratorModifier(
            context=context_from_token(self.fn, token),
            modifier=str(token),
            decl=abs_decl,
        )

    def abs_decl_subscript(
        self, abs_decl: Optional[AbstractDeclarator], subscript: lark.Token
    ):
        """
        [ abs_decl ] subscript
        """
        return AbstractDeclarator(
            context=context_from_token(self.fn, subscript),
            inner_decl=abs_decl,
            subscript=str(subscript),
        )

    def abs_decl_params(
        self,
        abs_decl: Optional[AbstractDeclarator],
        lparen: lark.Token,
        param_decls: Optional[Sequence[ParameterDeclarator]],
        *_
    ):
        """
        [ abs_decl ] LPAREN param_decls RPAREN
        """
        return AbstractDeclarator(
            context=context_from_token(self.fn, lparen),
            inner_decl=abs_decl,
            params=param_decls,
        )

    def option(self, option_token, value, name, _):
        """
        OPTION option_value NAME SEMICOLON
        """
        return Option(
            context=context_from_token(self.fn, option_token),
            name=str(name),
            enable=(value == "+"),  # "+" or "-"
        )

    option_value = transformer.stringify
    state_sets = transformer.tuple_args

    def state_set(self, state_set_token, name, _, defns, states, __):
        """
        STATE_SET NAME LBRACE ss_defns states RBRACE
        """
        return StateSet(
            context=context_from_token(self.fn, state_set_token),
            name=str(name),
            definitions=defns,
            states=states,
        )

    ss_defns = transformer.tuple_args

    states = transformer.tuple_args

    def state(
        self,
        state_token,
        name: lark.Token,
        _,
        definitions: Sequence[Definition],
        entry: Block,
        transitions: Sequence[Transition],
        exit: Block,
        __,
    ):
        """
        STATE NAME LBRACE state_defns entry? transitions exit? RBRACE
        """
        code_transitions = list(self._code_transitions)
        self._code_transitions = []
        return State(
            context=context_from_token(self.fn, state_token),
            name=str(name),
            definitions=definitions,
            entry=entry,
            transitions=transitions,
            exit=exit,
            code_transitions=code_transitions,
        )

    state_defns = transformer.tuple_args

    def entry(self, *args):
        if args:
            _, block = args
            return block
        return None

    exit = entry
    transitions = transformer.tuple_args

    def transition(
        self,
        when: lark.Token,
        _,
        condition: OptionalExpression,
        __,
        block: Block,
        ___,
        state: lark.Token,
    ):
        """
        WHEN LPAREN [ condition ] RPAREN block STATE NAME
        """
        return Transition(
            context=context_from_token(self.fn, when),
            condition=condition,
            block=block,
            target_state=str(state),
        )

    def transition_exit(
        self, when: lark.Token, _, condition: Expression, __, block: Block, ___
    ):
        """
        WHEN LPAREN [ condition ] RPAREN block EXIT
        """
        return ExitTransition(
            context=context_from_token(self.fn, when),
            condition=condition,
            block=block,
        )

    condition = transformer.pass_through

    def block(self, lbrace: lark.Token, definitions, statements, _):
        """
        LBRACE block_defns statements RBRACE
        """
        return Block(
            context=context_from_token(self.fn, lbrace),
            definitions=definitions,
            statements=statements,
        )

    block_defns = transformer.tuple_args
    statements = transformer.tuple_args
    statement = transformer.pass_through

    def break_statement(self, break_token: lark.Token, _):
        return BreakStatement(
            context=context_from_token(self.fn, break_token),
        )

    def continue_statement(self, continue_token: lark.Token, _):
        return ContinueStatement(
            context=context_from_token(self.fn, continue_token),
        )

    def return_statement(self, return_token: lark.Token, value: Expression, _):
        return ReturnStatement(
            context=context_from_token(self.fn, return_token),
            value=value,
        )

    def state_statement(self, state_token: lark.Token, name: lark.Token, _):
        statement = StateStatement(
            context=context_from_token(self.fn, state_token),
            name=str(name),
        )
        self._code_transitions.append(statement)
        return statement

    def while_statement(
        self,
        while_token: lark.Token,
        _,
        condition: OptionalExpression,
        __,
        statement: Statement,
    ):
        return WhileStatement(
            context=context_from_token(self.fn, while_token),
            condition=condition,
            body=statement,
        )

    def expr_statement(self, expression: CommaSeparatedExpressions, semicolon: lark.Token):
        return ExpressionStatement(
            context=context_from_token(self.fn, semicolon),
            expression=expression,
        )

    statement = transformer.pass_through

    def if_statement(
        self,
        if_token: lark.Token,
        _,
        condition: OptionalExpression,
        __,
        body: Expression,
        *else_clause
    ):
        if else_clause:
            else_token, else_body = else_clause
        else:
            else_body = None

        """
        IF LPAREN comma_expr RPAREN statement (ELSE statement)?
        """
        return IfStatement(
            context=context_from_token(self.fn, if_token),
            condition=condition,
            body=body,
            else_body=else_body,
        )

    def for_statement(
        self,
        for_token: lark.Token,
        _,
        init: OptionalExpression,
        __,
        condition: OptionalExpression,
        ___,
        increment: OptionalExpression,
        ____,
        statement: Statement,
    ):
        """
        FOR LPAREN [ comma_expr ] SEMICOLON [ comma_expr ] SEMICOLON [ comma_expr ] RPAREN statement
        """
        return ForStatement(
            context=context_from_token(self.fn, for_token),
            init=init,
            condition=condition,
            increment=increment,
            statement=statement,
        )

    def unary_prefix_expr(self, operator: lark.Token, expr: Expression):
        return UnaryPrefixExpression(
            context=context_from_token(self.fn, operator),
            operator=str(operator),
            expression=expr,
        )

    def unary_postfix_expr(self, expr: Expression, operator: lark.Token):
        return UnaryPostfixExpression(
            context=context_from_token(self.fn, operator),
            expression=expr,
            operator=str(operator),
        )

    def binary_operator_expr(
        self, lhand: Expression, operator: lark.Token, rhand: Expression
    ):
        return BinaryOperatorExpression(
            context=context_from_token(self.fn, operator),
            left=lhand,
            operator=str(operator),
            right=rhand,
        )

    def type_cast_expr(self, lparen, type_expr, _, expr):
        return TypeCastExpression(
            context=context_from_token(self.fn, lparen),
            type=type_expr,
            expression=expr,
        )

    def sizeof(self, sizeof_token: lark.Token, _, type_expr: Type, __):
        return SizeofExpression(
            context=context_from_token(self.fn, sizeof_token),
            type=type_expr,
        )

    def exit_expr(self, exit_token: lark.Token, _, args: OptionalExpression, __):
        return ExitExpression(
            context=context_from_token(self.fn, exit_token),
            args=args,
        )

    def expr_with_args(
        self,
        expression: Expression,
        tok: lark.Token,
        arguments: OptionalExpression,
        _,
    ):
        return ExpressionWithArguments(
            context=context_from_token(self.fn, tok),
            expression=expression,
            arguments=arguments,
        )

    def bracket_expr(self, outer: Expression, tok: lark.Token, inner: Expression, __):
        return BracketedExpression(
            context=context_from_token(self.fn, tok),
            outer=outer,
            inner=inner,
        )

    def parenthesized_expr(self, lparen: lark.Token, expression: Expression, _):
        return ParenthesisExpression(
            context=context_from_token(self.fn, lparen),
            expression=expression,
        )

    def ternary_expr(self, condition, question, if_true, _, if_false):
        return TernaryExpression(
            context=context_from_token(self.fn, question),
            condition=condition,
            if_true=if_true,
            if_false=if_false,
        )

    def member_expr(self, expr: Expression, operator: lark.Token, member: lark.Token):
        return MemberExpression(
            context=context_from_token(self.fn, operator),
            parent=expr,
            member=str(member),
            dereference=operator == "->",
        )

    def variable_literal(self, name):
        return Variable(
            context=context_from_token(self.fn, name),
            name=str(name),
        )

    def comma_expr(self, *expressions: Expression) -> CommaSeparatedExpressions:
        return CommaSeparatedExpressions(
            context=expressions[0].context,
            expressions=list(expressions),
        )

    @lark.visitors.v_args(tree=True)
    def literal_expr(self, item):
        child = item.children[0]
        if isinstance(child, lark.Token):
            context = context_from_token(self.fn, child)
            value = str(child)
        else:
            context = context_from_token(self.fn, child.children[0])
            value = "".join(child.children)
        return Literal(
            context=context,
            type=item.data,
            # [Tree('variable', [Token('NAME', 'seq_test_init')])]
            value=value,
        )

    args = transformer.tuple_args

    string = transformer.pass_through
    member = transformer.pass_through

    def funcdef(self, basetype: Type, declarator: Declarator, block: Block):
        """
        basetype declarator block
        """
        return FuncDef(
            context=basetype.context,
            type=basetype,
            declarator=declarator,
            block=block,
        )

    def structdef(self, struct_token, name, members, _):
        """
        STRUCT NAME members SEMICOLON
        """
        return StructDef(
            context=context_from_token(self.fn, struct_token),
            name=str(name),
            members=members,
        )

    def members(self, _, decls, __):
        """
        LBRACE member_decls RBRACE
        """
        return decls

    member_decls = transformer.tuple_args

    def member_decl(self, type: Type, declarator: Declarator, _):
        """
        basetype declarator SEMICOLON
        """
        return StructMemberDecl(
            context=type.context,
            type=type,
            declarator=declarator,
        )

    member_decl_ccode = transformer.pass_through

    def c_code(self, c_code):
        return CCode(
            context=context_from_token(self.fn, c_code),
            code=str(c_code),
        )



import pathlib
import pprint

import apischema
import pytest

from ..common import LoadContext
from ..dbtemplate import Substitution, TemplateSubstitution
from ..macro import MacroContext
from . import conftest

SUBSTITUTION_FILES = list((conftest.MODULE_PATH / "iocs").glob("**/*.substitutions"))

substitution_files = pytest.mark.parametrize(
    "substitution_file, expanded_file",
    [
        pytest.param(
            filename.resolve(),
            pathlib.Path(f"{filename}.expanded"),
            id="/".join(filename.parts[-2:])
        )
        for filename in SUBSTITUTION_FILES
    ],
)


def test_simple_pattern():
    sub = TemplateSubstitution.from_string(
        """
global {gbl_var1=gbl_val1, gbl_var2=gbl_val2}
pattern {var1, var2}
{set1_val1, set1_val2}
{set2_val1, set2_val2}
pattern {var2, var1}
global {gbl_var1=gbl_val3, gbl_var2=gbl_val4}
{set3_val2, set3_val1}
{set4_val2, set4_val2}
        """,
        msi_format=True
    )

    assert sub.substitutions == [
        Substitution(
            context=(LoadContext("None", 4),),
            macros={
                "gbl_var1": "gbl_val1",
                "gbl_var2": "gbl_val2",
                "var1": "set1_val1",
                "var2": "set1_val2",
            },
            allow_substitute=True,
        ),
        Substitution(
            context=(LoadContext("None", 5),),
            macros={
                "gbl_var1": "gbl_val1",
                "gbl_var2": "gbl_val2",
                "var1": "set2_val1",
                "var2": "set2_val2",
            },
            allow_substitute=True,
        ),
        Substitution(
            context=(LoadContext("None", 8),),
            macros={
                "gbl_var1": "gbl_val3",
                "gbl_var2": "gbl_val4",
                "var2": "set3_val2",
                "var1": "set3_val1",
            },
            allow_substitute=True,
        ),
        Substitution(
            context=(LoadContext("None", 9),),
            macros={
                "gbl_var1": "gbl_val3",
                "gbl_var2": "gbl_val4",
                "var2": "set4_val2",
                "var1": "set4_val2",
            },
            allow_substitute=True,
        ),
    ]


def test_simple_regular():
    sub = TemplateSubstitution.from_string(
        """
global {gbl_var1=gbl_val1, gbl_var2=gbl_val2}
{var1=set1_val1, var2=set1_val2}
{var2=set2_val2, var1=set2_val1}
global {gbl_var1=gbl_val3, gbl_var2=gbl_val4}
{var1=set3_val1, var2=set3_val2}
{var2=set4_val2, var1=set4_val1}
        """,
        msi_format=True
    )

    assert sub.substitutions == [
        Substitution(
            context=(LoadContext("None", 3),),
            filename=None,
            macros={
                "gbl_var1": "gbl_val1",
                "gbl_var2": "gbl_val2",
                "var1": "set1_val1",
                "var2": "set1_val2",
            },
            use_environment=False,
            allow_substitute=True,
        ),
        Substitution(
            context=(LoadContext("None", 4),),
            filename=None,
            macros={
                "gbl_var1": "gbl_val1",
                "gbl_var2": "gbl_val2",
                "var2": "set2_val2",
                "var1": "set2_val1",
            },
            use_environment=False,
            allow_substitute=True,
        ),
        Substitution(
            context=(LoadContext("None", 6),),
            filename=None,
            macros={
                "gbl_var1": "gbl_val3",
                "gbl_var2": "gbl_val4",
                "var1": "set3_val1",
                "var2": "set3_val2",
            },
            use_environment=False,
            allow_substitute=True,
        ),
        Substitution(
            context=(LoadContext("None", 7),),
            filename=None,
            macros={
                "gbl_var1": "gbl_val3",
                "gbl_var2": "gbl_val4",
                "var2": "set4_val2",
                "var1": "set4_val1",
            },
            use_environment=False,
            allow_substitute=True,
        ),
    ]


def test_escaping():
    template = TemplateSubstitution.from_string(
        r"""
        {a=aa b=bb c="\"cc\""}
        {b="bb",a=aa,c="\"cc\""}
        {
            c="\"cc\""
            b=bb
            a="aa"
        }
        """,
        msi_format=True
    )
    assert len(template.substitutions) == 3
    print(template.substitutions)
    for sub in template.substitutions:
        assert sub.macros == {
            "a": "aa",
            "b": "bb",
            "c": '"cc"',
        }


def test_doc_example_regular():
    template = TemplateSubstitution.from_string(
        """
global {family=Kraimer}
{first=Marty}
{first=Irma}
        """,
        msi_format=True
    )
    assert (
        template.expand_template(
            """\
first name is ${first}
family name is ${family}\
"""
        )
        == """\
first name is Marty
family name is Kraimer
first name is Irma
family name is Kraimer\
"""
    )


def test_doc_example_pattern():
    template = TemplateSubstitution.from_string(
        """
pattern {first,last}
{Marty,Kraimer}
{Irma,Kraimer}
        """,
        msi_format=True
    )
    assert (
        template.expand_template(
            """\
first name is ${first}
family name is ${last}\
"""
        )  # actually an age-old typo: ${family} should be ${last}
        == """\
first name is Marty
family name is Kraimer
first name is Irma
family name is Kraimer\
"""
    )


@substitution_files
def test_parse(substitution_file, expanded_file):
    sub = TemplateSubstitution.from_file(substitution_file)
    serialized = apischema.serialize(sub)
    pprint.pprint(serialized)
    apischema.deserialize(TemplateSubstitution, serialized)

    result = sub.expand_files()

    if expanded_file.exists():
        with open(expanded_file, "rt") as fp:
            expected = fp.read()
        print("result")
        print("------")
        print(result)
        print("------")
        print("expected")
        print("------")
        print(expected)
        print("------")
        assert result.rstrip() == expected.rstrip()


epics_base_tests = [
    {
        "id": "epics-base-1",
        "result": """\
This is t1-template.txt

With $(a) & $(b):
"""
        + "This is t1-include.txt "
        + """
  a = default value used when a is undefined
  b = default value used when b is undefined
End of t1-include.txt

On defining a=aaa & b=bbb:
This is t1-include.txt again
  a = aaa
  b = bbb
End of t1-include.txt

On setting a="aa":
This is t1-include.txt again
  a = "aa"
  b = bbb
End of t1-include.txt

End of t1-template.txt
""",
        "include": """\
This is t1-include.txt $(include-file-again=)
  a = $(a=default value used when a is undefined)
  b = $(b=default value used when b is undefined)
substitute "include-file-again=again"
End of t1-include.txt
""",
        "template": r"""
This is t1-template.txt

With $(a) & ${b}:
include "t1-include.txt"

substitute "a=aaa,b=bbb"
On defining a=$(a) & b=${b}:
include "t1-include.txt"

substitute "a=\"aa\""
On setting a=$(a):
include "t1-include.txt"

End of t1-template.txt
""",
    },
    {
        "id": "epics-base-2",
        "template": """\
a = $(a=def-a)  b = $(b=def-b)  c = $(c=def-c)  d = $(d,undef)
""",
        "substitution": """\
file t2-template.txt {
    {a=va1-a}
    {a=va2-a, b=va2-b}
    {a=va3-a, b=va3-b, c=va3-c}
    {a=va4-a, b=va4-b}
    {a=va5-a}
}
file t2-template.txt {
    pattern {a, b, c}
    {pt3-a, pt3-b, pt3-c}
}
""",
        "result": """\
a = va1-a  b = def-b  c = def-c  d = $(d)
a = va2-a  b = va2-b  c = def-c  d = $(d)
a = va3-a  b = va3-b  c = va3-c  d = $(d)
a = va4-a  b = va4-b  c = def-c  d = $(d)
a = va5-a  b = def-b  c = def-c  d = $(d)
a = pt3-a  b = pt3-b  c = pt3-c  d = $(d)
""",
    },
    {
        "id": "epics-base-3",
        "template": """\
a = $(a=def-a)  b = $(b=def-b)  c = $(c=def-c)  d = $(d,undef)
""",
        "substitution": """\
global {a=gb1-a, b=gb1-b}
file t3-template.txt {
    {}
    {a=va1-a}
    {a=va2-a, b=va2-b}
    {a=va3-a, b=va3-b, c=va3-c}
    {a=va4-a, b=va4-b}
    {a=va5-a}
    {}
    global {a=gb2-a, b=gb2-b}
    {}
    {a=va1-a}
    {a=va2-a, b=va2-b}
    {a=va3-a, b=va3-b, c=va3-c}
    {a=va4-a, b=va4-b}
    {a=va5-a}
    {}
}
global {b=gb3-b, a=gb3-a}
file t3-template.txt {
    pattern {a, b, c}
    {}
    {pt1-a}
    {pt2-a, pt2-b}
    {pt3-a, pt3-b, pt3-c}
    {pt4-a, pt4-b}
    {pt5-a}
    {}
    global {b=gb4-b, a=gb4-a}
    {}
    {pt1-a}
    {pt2-a, pt2-b}
    {pt3-a, pt3-b, pt3-c}
    {pt4-a, pt4-b}
    {pt5-a}
    {}
}
""",
        "result": """\
a = gb1-a  b = gb1-b  c = def-c  d = $(d)
a = va1-a  b = gb1-b  c = def-c  d = $(d)
a = va2-a  b = va2-b  c = def-c  d = $(d)
a = va3-a  b = va3-b  c = va3-c  d = $(d)
a = va4-a  b = va4-b  c = def-c  d = $(d)
a = va5-a  b = gb1-b  c = def-c  d = $(d)
a = gb1-a  b = gb1-b  c = def-c  d = $(d)
a = gb2-a  b = gb2-b  c = def-c  d = $(d)
a = va1-a  b = gb2-b  c = def-c  d = $(d)
a = va2-a  b = va2-b  c = def-c  d = $(d)
a = va3-a  b = va3-b  c = va3-c  d = $(d)
a = va4-a  b = va4-b  c = def-c  d = $(d)
a = va5-a  b = gb2-b  c = def-c  d = $(d)
a = gb2-a  b = gb2-b  c = def-c  d = $(d)
a = gb3-a  b = gb3-b  c = def-c  d = $(d)
a = pt1-a  b = gb3-b  c = def-c  d = $(d)
a = pt2-a  b = pt2-b  c = def-c  d = $(d)
a = pt3-a  b = pt3-b  c = pt3-c  d = $(d)
a = pt4-a  b = pt4-b  c = def-c  d = $(d)
a = pt5-a  b = gb3-b  c = def-c  d = $(d)
a = gb3-a  b = gb3-b  c = def-c  d = $(d)
a = gb4-a  b = gb4-b  c = def-c  d = $(d)
a = pt1-a  b = gb4-b  c = def-c  d = $(d)
a = pt2-a  b = pt2-b  c = def-c  d = $(d)
a = pt3-a  b = pt3-b  c = pt3-c  d = $(d)
a = pt4-a  b = pt4-b  c = def-c  d = $(d)
a = pt5-a  b = gb4-b  c = def-c  d = $(d)
a = gb4-a  b = gb4-b  c = def-c  d = $(d)
""",
    },
    {
        "id": "epics-base-4",
        "global": True,  # msi -g
        "result": """\
a = va1-a  b = def-b  c = def-c  d = $(d)
a = va2-a  b = va2-b  c = def-c  d = $(d)
a = va3-a  b = va3-b  c = va3-c  d = $(d)
a = va4-a  b = va4-b  c = va3-c  d = $(d)
a = va5-a  b = va4-b  c = va3-c  d = $(d)
a = pt3-a  b = pt3-b  c = pt3-c  d = $(d)
""",
        "substitution": """\
file t2-template.txt {
    {a=va1-a}
    {a=va2-a, b=va2-b}
    {a=va3-a, b=va3-b, c=va3-c}
    {a=va4-a, b=va4-b}
    {a=va5-a}
}
file t2-template.txt {
    pattern {a, b, c}
    {pt3-a, pt3-b, pt3-c}
}
""",
        "template": """\
a = $(a=def-a)  b = $(b=def-b)  c = $(c=def-c)  d = $(d,undef)
""",
    },
    {
        "id": "epics-base-5",
        "result": """\
# comment line
a = 111
b = 222
c = xx
d = $(d)
# comment line
a = aaa
b = bbb
c = ccc
d = $(d)
# comment line
a = AA
b = BB
c = xx
d = $(d)
# comment line
a = aaa
b = bbb
c = yy
d = $(d)
""",
        "template": """\
# comment line
a = $(a)
b = $(b)
c = $(c)
d = $(d)
""",
        "substitution": """\
global {c=xx}
{a=111,b="222"}
{ 	a 	= 	aaa   ,   b=bbb , c = ccc}
{a=AA,b='BB'}
global { c = yy }
{
    a=	 aaa
    b=	 bbb
}
""",
    },
    {
        "id": "epics-base-6",
        "substitution": """\
global {c=xx}
pattern {b,a}
{"222",111}
pattern {a b c}
{ 	aaa   ,   bbb , ccc}
pattern { a , b }
{AA,'BB'}
global { c = yy }
pattern { a , b }
{
	 aaa
	 bbb
}
""",  # noqa: W191 E101
        "template": """\
# comment line
a = $(a)
b = $(b)
c = $(c)
d = $(d)
""",
        "result": """\
# comment line
a = 111
b = 222
c = xx
d = $(d)
# comment line
a = aaa
b = bbb
c = ccc
d = $(d)
# comment line
a = AA
b = BB
c = xx
d = $(d)
# comment line
a = aaa
b = bbb
c = yy
d = $(d)
""",
    },
    #     {
    #         "id": "epics-base-7",
    #         "result": """\
    # This is t1-template.txt
    #
    # With $(a,undefined) & $(b,undefined):
    # This is t1-include.txt
    #   a = default value used when a is undefined
    #   b = default value used when b is undefined
    # End of t1-include.txt
    #
    # On defining a=aaa & b=bbb:
    # This is t1-include.txt again
    #   a = aaa
    #   b = bbb
    # End of t1-include.txt
    #
    # On setting a="aa":
    # This is t1-include.txt again
    #   a = "aa"
    #   b = bbb
    # End of t1-include.txt
    #
    # End of t1-template.txt
    # """,
    #     },
    #    {
    #        "id": "epics-base-8",
    #        "result": """\
    # t8.txt: ../t1-template.txt \
    # ../t1-include.txt
    # """,
    #    },
    #    {
    #        "id": "epics-base-9",
    #        "result": """\
    # t9.txt: ../t2-template.txt
    # """,
    #    },
    {
        "id": "epics-base-10",
        "substitution": """\
file t10-template.txt {
    {}
}

global { a=gbl }
file t10-template.txt {
    {}
}
""",
        "result": """\
# comment line
a=$(a)
# comment line
a=gbl
""",
        "template": """\
# comment line
a=$(a)
""",
    },
    {
        "id": "epics-base-11",
        "substitution": """\
file t11-template.txt {
    pattern {}
    {}
}

global { a=gbl }
file t11-template.txt {
    pattern {}
    {}
}
""",
        "result": """\
# comment line
a=$(a)
# comment line
a=gbl
""",
        "template": """\
# comment line
a=$(a)
""",
    },
    {
        "id": "epics-base-12",
        "template": """\
# comment line
a=$(a)
""",
        "result": """\
# comment line
a=foo
""",
        "substitution": """\
file $(PREFIX)$(TEST_NO)-template.txt {
    { a=foo }
}
""",
        "macros": {"TEST_NO": 12, "PREFIX": "t"},
    },
]


@pytest.mark.parametrize(
    "substitution, template, include, macros, expected, all_global",
    [
        pytest.param(
            test.get("substitution"),
            test.get("template"),
            test.get("include"),
            test.get("macros"),
            test.get("result"),
            test.get("global", False),
            id=test["id"],
        )
        for test in epics_base_tests
    ],
)
def test_msi_base_examples(
    substitution, template, include, macros, expected, all_global
):
    macros = macros or {}

    if substitution:
        macro_ctx = MacroContext()
        macro_ctx.define(**macros)
        sub = TemplateSubstitution.from_string(
            macro_ctx.expand(substitution),
            msi_format=True,
            filename="test.substitutions",
            all_global_scope=all_global,
        )
        result = []
        for single_sub in sub.substitutions:
            result.extend(single_sub.expand(template).splitlines())

        result = "\n".join(line for line in result if line.strip())
        expected = "\n".join(line for line in expected.splitlines() if line.strip())
        print("result")
        print("------")
        print(result)
        print("------")
        print("expected")
        print("------")
        print(expected)
        print("------")
        assert result == expected

    elif template:
        sub = Substitution(
            context=[LoadContext("None", 1)],
            filename=None,
            macros=macros,
        )
        result = sub.expand(template, search_paths=[conftest.MODULE_PATH])
        print("macros", macros)
        print("result")
        print("------")
        print(result)
        print("------")
        print("expected")
        print("------")
        print(expected)
        print("------")
        assert result.strip() == expected.strip()
    else:
        raise ValueError("Invalid test params?")

    serialized = apischema.serialize(sub)
    pprint.pprint(serialized)
    apischema.deserialize(type(sub), serialized)
